from verry.interval.floatinterval import FloatConverter
from verry.interval.interval import RoundingMode
from verry.misc.formatspec import FormatSpec


def test_converter():
    ROUND_CEILING = RoundingMode.ROUND_CEILING
    ROUND_FLOOR = RoundingMode.ROUND_FLOOR
    converter = FloatConverter()

    assert converter.format(0.001, FormatSpec(".3f"), ROUND_CEILING) == "0.002"
    assert converter.format(0.001, FormatSpec(".3f"), ROUND_FLOOR) == "0.001"
    assert converter.format(-1.23, FormatSpec(".3g"), ROUND_CEILING) == "-1.22"
    assert converter.format(-1.23, FormatSpec(".3g"), ROUND_FLOOR) == "-1.23"
    assert converter.format(-12.1, FormatSpec(".3e"), ROUND_CEILING) == "-1.209e+1"
    assert converter.format(-12.1, FormatSpec(".3e"), ROUND_FLOOR) == "-1.210e+1"



"""
####################################################
Quadrature and solving ODEs (:mod:`verry.integrate`)
####################################################

.. currentmodule:: verry.integrate

This module provides verified routines for quadrature and solving ordinary differential
equations.

Quadrature functions
====================

.. autosummary::
    :toctree: generated/

    cumulative_simpson
    cumulative_trapezoid

Solving initial value problems for ODEs
=======================================

Solvers
-------

.. autosummary::
    :toctree: generated/

    C0Solver
    C1Solver

Integrators
-----------

.. autosummary::
    :toctree: generated/

    Integrator
    eilo
    kashi

Trackers
--------

.. autosummary::
    :toctree: generated/

    Tracker
    affinetracker
    directtracker
    doubletontracker
    qrtracker

Miscellaneous
-------------

.. autosummary::
    :toctree: generated/

    AbortSolving
    C0SolverCallbackArg
    C0SolverResultContent
    C1SolverCallbackArg
    C1SolverResultContent
    OdeSolution
    SolverResult
    seriessolution
    variationaleq

"""

from .integrator import Integrator, eilo, kashi
from .quad import cumulative_simpson, cumulative_trapezoid
from .solver import (
    AbortSolving,
    C0Solver,
    C0SolverCallbackArg,
    C0SolverResultContent,
    C1Solver,
    C1SolverCallbackArg,
    C1SolverResultContent,
    OdeSolution,
    SolverResult,
)
from .tracker import Tracker, affinetracker, directtracker, doubletontracker, qrtracker
from .utility import seriessolution, variationaleq

__all__ = [
    "Integrator",
    "eilo",
    "kashi",
    "cumulative_simpson",
    "cumulative_trapezoid",
    "AbortSolving",
    "C0Solver",
    "C0SolverCallbackArg",
    "C0SolverResultContent",
    "C1Solver",
    "C1SolverCallbackArg",
    "C1SolverResultContent",
    "OdeSolution",
    "SolverResult",
    "Tracker",
    "affinetracker",
    "directtracker",
    "doubletontracker",
    "qrtracker",
    "seriessolution",
    "variationaleq",
]



"""
###########################################
Interval arithmetic (:mod:`verry.interval`)
###########################################

.. currentmodule:: verry.interval

This module provides basic interval arithmetic.

Intervals
=========

.. autosummary::
    :toctree: generated/

    Interval
    FloatInterval

Miscellaneous
=============

.. autosummary::
    :toctree: generated/

    Converter
    Operator
    RoundingMode

"""

from .floatinterval import FloatInterval
from .interval import (
    Converter,
    Interval,
    Operator,
    RoundingMode,
)

__all__ = [
    "FloatInterval",
    "Converter",
    "Interval",
    "Operator",
    "RoundingMode",
]



import decimal
import fractions
import math
import re
import sys
from typing import assert_never

from verry import function as vrf
from verry.interval import _floatoperator  # type: ignore
from verry.interval.interval import Converter, Interval, Operator, RoundingMode
from verry.misc.formatspec import FormatSpec


def _decimal_exponent(x: decimal.Decimal | fractions.Fraction) -> int:
    if x == 0:
        return 0

    result = 0

    while abs(x) < 1:
        x *= 10
        result -= 1

    while abs(x) >= 10:
        x /= 10
        result += 1

    return result


class FloatConverter(Converter[float]):
    __slots__ = ()

    def fromfloat(self, value, strict=True):
        return value

    def fromstr(self, value, rounding) -> float:
        if rounding == RoundingMode.ROUND_FAST:
            return float(value)

        if re.fullmatch("[-+]?inf(?:inity)?", value, re.I) is not None:
            return -math.inf if value[0] == "-" else math.inf

        try:
            frac = fractions.Fraction(value)
        except ValueError:
            raise ValueError(f"Could not convert string to float: '{value}'")

        if frac == 0:
            return 0.0

        if is_negative := frac < 0:
            frac *= -1

        if frac < sys.float_info.min:
            match rounding:
                case RoundingMode.ROUND_CEILING:
                    return 0.0 if is_negative else -sys.float_info.min

                case RoundingMode.ROUND_FLOOR:
                    return -sys.float_info.min if is_negative else 0.0

        if frac > sys.float_info.max:
            match rounding:
                case RoundingMode.ROUND_CEILING:
                    return -sys.float_info.max if is_negative else math.inf

                case RoundingMode.ROUND_FLOOR:
                    return -math.inf if is_negative else sys.float_info.max

        shift = 0

        while frac < 2**52:
            frac *= 2
            shift += 1

        while frac >= 2**53:
            frac /= 2
            shift -= 1

        div, mod = divmod(frac.numerator, frac.denominator)
        tmp = float(div)

        if is_negative:
            tmp *= -1.0

        if shift > 0:
            for _ in range(shift):
                tmp *= 0.5
        else:
            for _ in range(-shift):
                tmp *= 2.0

        if mod == 0:
            return tmp

        match rounding:
            case RoundingMode.ROUND_CEILING:
                return math.nextafter(tmp, float("inf"))

            case RoundingMode.ROUND_FLOOR:
                return math.nextafter(tmp, -float("inf"))

            case _ as unreachable:
                assert_never(unreachable)

    def fromint(self, value, rounding) -> float:
        if abs(value) <= 0x1FFFFFFFFFFFFF:
            return float(value)

        return self.fromstr(str(value), rounding)

    def tostr(self, value, mode):
        return self.format(value, FormatSpec(), mode)

    def format(self, value, spec, rounding):
        if value == 0.0 or not math.isfinite(value):
            return spec.format(value)

        context = decimal.Context(Emin=decimal.MIN_EMIN, Emax=decimal.MAX_EMAX)
        frac = fractions.Fraction(value)
        prec = spec.prec

        if prec is None:
            prec = 6

        match spec.type:
            case "e" | "E":
                prec_pow = 10**prec
                shift = 0

                while abs(frac) < prec_pow:
                    frac *= 10
                    shift += 1

                while abs(frac) >= 10 * prec_pow:
                    frac /= 10
                    shift -= 1

                with decimal.localcontext(context, prec=prec + 1):
                    match rounding:
                        case RoundingMode.ROUND_CEILING:
                            tmp = decimal.Decimal(math.ceil(frac))

                        case RoundingMode.ROUND_FLOOR:
                            tmp = decimal.Decimal(math.floor(frac))

                    if shift > 0:
                        for _ in range(shift):
                            tmp /= 10
                    else:
                        for _ in range(-shift):
                            tmp *= 10

                    return spec.format(tmp)

            case "f" | "F":
                exp = _decimal_exponent(frac)

                for _ in range(prec):
                    frac *= 10

                with decimal.localcontext(context, prec=max(2, prec + exp + 1)):
                    match rounding:
                        case RoundingMode.ROUND_CEILING:
                            tmp = decimal.Decimal(math.ceil(frac))

                        case RoundingMode.ROUND_FLOOR:
                            tmp = decimal.Decimal(math.floor(frac))

                    for _ in range(prec):
                        tmp /= 10

                    return spec.format(tmp)

            case "g" | "G" | None:
                exp = _decimal_exponent(round(frac, prec))

                if not -4 <= exp < prec:
                    spec = spec.replace(prec=prec - 1, type="e")
                    result = self.format(value, spec, rounding)
                    return result

                spec = spec.replace(prec=prec - 1 - exp, type="f")
                result = self.format(value, spec, rounding)

                while "." in result and result[-1] in {".", "0"}:
                    result = result[:-1]

                return result

    def repr(self, value):
        if not math.isfinite(value):
            return repr(value)

        return f"<{value.hex()}>"


class FloatOperator(Operator[float]):
    __slots__ = ()
    ZERO = 0.0
    ONE = 1.0
    INFINITY = math.inf

    def cadd(self, lhs, rhs):
        return _floatoperator.cadd(lhs, rhs)

    def cmul(self, lhs, rhs):
        return _floatoperator.cmul(lhs, rhs)

    def cdiv(self, lhs, rhs):
        return _floatoperator.cdiv(lhs, rhs)

    def csqr(self, value):
        return _floatoperator.csqr(value)

    def fsqr(self, value):
        return _floatoperator.fsqr(value)


class FloatInterval(Interval[float]):
    """Double-precision inf-sup type interval.

    Parameters
    ----------
    inf : float | int | str | None, optional
        Infimum of the interval.
    sup : float | int | str | None, optional
        Supremum of the interval.

    Attributes
    ----------
    inf : float
        Infimum of the interval.
    sup : float
        Supremum of the interval.
    converter : Converter
    endtype : type[float]
    operator : Operator
    """

    __slots__ = ()
    converter = FloatConverter()
    operator = FloatOperator()
    endtype = float

    def mid(self) -> float:
        ZERO = self.operator.ZERO
        INFINITY = self.operator.INFINITY

        if self.inf == -INFINITY:
            return ZERO if self.sup == INFINITY else self.sup

        if self.sup == INFINITY:
            return self.inf

        if abs(self.inf) >= 1 and abs(self.sup) >= 1:
            return self.inf / 2 + self.sup / 2

        return (self.inf + self.sup) / 2

    @classmethod
    def __exp_point(cls, x):
        SQRTE_INV_RD = float.fromhex("0x1.368b2fc6f9609p-1")
        SQRTE_RU = float.fromhex("0x1.a61298e1e069cp+0")
        h = cls(x) - round(x)
        r = vrf.e(h) ** round(x)
        a = cls(0.0)
        tmp = cls(1.0)

        for i in range(1, 16):
            a += tmp
            tmp *= h / i

        a += cls(SQRTE_INV_RD, SQRTE_RU) * tmp
        r *= a
        return r

    @classmethod
    def __log_point(cls, x):
        TWO_THIRDS_RU = float.fromhex("0x1.5555555555556p-1")
        ERROR = float.fromhex("0x1.973774dfc4858p+3")
        p = 0

        while x < TWO_THIRDS_RU:
            x *= 2.0
            p += 1

        while x > 2 * TWO_THIRDS_RU:
            x /= 2.0
            p -= 1

        u = cls(x)
        y = (u - 1.0) / (u + 1.0)
        r = -p * vrf.ln2(u) + 2.0 * y
        tmp = y

        for k in range(2, 14):
            tmp *= y**2
            r += 2.0 * tmp / (2.0 * k - 1.0)

        tmp *= y
        r += cls(-ERROR, ERROR) * tmp
        return r

    def _verry_overload_(self, fun, *args, **kwargs):
        if fun is vrf.e:
            E_INF = float.fromhex("0x1.5bf0a8b145769p+1")
            E_SUP = float.fromhex("0x1.5bf0a8b14576ap+1")
            return self.__class__(E_INF, E_SUP)

        if fun is vrf.exp:
            return self.__exp()

        if fun is vrf.ln2:
            LN2_INF = float.fromhex("0x1.62e42fefa39efp-1")
            LN2_SUP = float.fromhex("0x1.62e42fefa39f0p-1")
            return self.__class__(LN2_INF, LN2_SUP)

        if fun is vrf.log:
            return self.__log()

        if fun is vrf.pi:
            PI_INF = float.fromhex("0x1.921fb54442d18p+1")
            PI_SUP = float.fromhex("0x1.921fb54442d19p+1")
            return self.__class__(PI_INF, PI_SUP)

        if fun is vrf.pow:
            return vrf.exp(vrf.log(self.ensure(args[0])) * args[1])

        return super()._verry_overload_(fun, *args, **kwargs)

    def __exp(self):
        inf = self.__exp_point(self.inf).inf if math.isfinite(self.inf) else 0.0
        sup = self.__exp_point(self.sup).sup if math.isfinite(self.sup) else math.inf
        return self.__class__(inf, sup)

    def __log(self):
        if self.inf < 0.0:
            raise ValueError("math domain error")

        inf = self.__log_point(self.inf).inf if self.inf != 0.0 else -math.inf
        sup = self.__log_point(self.sup).sup if math.isfinite(self.sup) else math.inf
        return self.__class__(inf, sup)



import itertools
from collections.abc import Callable
from typing import Any

import numpy as np

from verry.autodiff.autodiff import deriv, grad
from verry.interval.interval import Interval
from verry.linalg.intervalmatrix import IntervalMatrix


def branchbound[T: Interval](
    fun: Callable,
    domain: IntervalMatrix[T],
    fprime: Callable | bool = True,
    xtol: Any = float("inf"),
    ytol: Any = 0.0,
    max_iter: int = 16,
) -> tuple[T, list[IntervalMatrix[T]]]:
    """Find the minimum of the multivariate scalar-valued function by repeatedly
    dividing the interval vector.

    Parameters
    ----------
    fun : Callable
        Function to be optimized.
    domain : IntervalMatrix
        Interval vector in which to search for the minimum value.
    fprime : Callable | bool, optional
        Gradient of `fun` (the default is ``grad(fun)``). If `fprime` is ``False``,
        The differentiability of `fun` is not assumed.
    xtol : default=inf
        Absolute tolerance.
    ytol : default=0.0
        Relative tolerance.
    max_iter : int, default=16
        Maximum number of iterations.

    Returns
    -------
    r0 : Interval
        Interval containing the minimum value.
    r1 : list[IntervalMatrix]
        Interval vectors at which `fun` may take a minimum value.

    Warnings
    --------
    `fun` must be a :math:`C^1`-function on `domain` if `fprime` is not ``False``.
    Futhermore, `fun` must neither be a constant nor contain conditional branches
    (cf. :doc:`/userguide/pitfall`).

    See Also
    --------
    verry.autodiff.grad

    Examples
    --------
    >>> from verry.linalg import FloatIntervalMatrix as FIM
    >>> y, x = branchbound(lambda x, y: x**2 + y, FIM(inf=[-2, -1], sup=[1, 2]))
    >>> print(format(y, ".3f"))
    [-1.000, -0.999]
    """
    if not isinstance(domain, IntervalMatrix):
        raise TypeError

    intvlmat = type(domain)
    intvl = domain.interval
    ZERO = intvl.operator.ZERO
    INFINITY = intvl.operator.INFINITY

    if fprime is True:
        fprime = grad(fun)

    if isinstance(xtol, (float, int)):
        xtol = intvl.converter.fromfloat(float(xtol), strict=False)

    if not isinstance(xtol, intvl.endtype):
        raise TypeError

    if isinstance(ytol, (float, int)):
        ytol = intvl.converter.fromfloat(float(ytol), strict=False)

    if not isinstance(ytol, intvl.endtype):
        raise TypeError

    if max_iter <= 0:
        raise ValueError

    branches = [domain]
    bound = inf = sup = INFINITY
    argmins: list[IntervalMatrix[T]] = []

    for i in range(max_iter):
        if not branches:
            break

        next_branches: list[IntervalMatrix[T]] = []

        for x in branches:
            y = fun(*x)

            if y.inf > bound:
                continue

            x0 = x.mid()
            y0 = fun(*intvlmat(x0, intvl=intvl))

            if fprime:
                c = intvlmat(fprime(*x), intvl=intvl)

                if np.all((c.inf > ZERO) & (x.inf != domain.inf)):
                    continue

                if np.all((c.sup < ZERO) & (x.sup != domain.sup)):
                    continue

                y &= y0 + c @ (x - x0)

                if y.inf > bound:
                    continue

            bound = min(bound, y0.sup)
            is_sufficient = (x.diam() <= xtol) & (y.diam() <= abs(y.mid()) * ytol)

            if np.all(is_sufficient) or i == max_iter - 1:
                argmins.append(x)
                inf = min(inf, y.inf)
                sup = min(sup, y.sup)
                continue

            tmp = ((intvl(y.inf, y.mid()), intvl(y.mid(), y.sup)) for y in x)
            next_branches.extend(
                intvlmat(x, intvl=intvl) for x in itertools.product(*tmp)
            )

        branches = next_branches

    return intvl(inf, sup), argmins


def branchbound_scalar[T: Interval](
    fun: Callable,
    domain: T,
    fprime: Callable | bool = True,
    xtol: Any = float("inf"),
    ytol: Any = 0.0,
    max_iter: int = 16,
) -> tuple[T, list[T]]:
    """Find the minimum of the univariate scalar-valued function by repeatedly
    dividing the interval.

    Parameters
    ----------
    fun : Callable
        Function to be optimized.
    domain : Interval
        Interval in which to search for the minimum value.
    fprime : Callable | bool, optional
        Derivative of `fun` (the default is ``deriv(fun)``). If `fprime` is ``False``,
        The differentiability of `fun` is not assumed.
    xtol : default=inf
        Absolute tolerance.
    ytol : default=0.0
        Relative tolerance.
    max_iter : int, default=16
        Maximum number of iterations.

    Returns
    -------
    r0 : Interval
        Interval containing the minimum value.
    r1 : list[Interval]
        Intervals at which `fun` may take a minimum value.

    Warnings
    --------
    `fun` must be a :math:`C^1`-function on `domain` if `fprime` is not ``False``.
    Futhermore, `fun` must neither be a constant nor contain conditional branches
    (cf. :doc:`/userguide/pitfall`).

    See Also
    --------
    verry.autodiff.deriv

    Examples
    --------
    >>> from verry import FloatInterval as FI
    >>> y, x = branchbound_scalar(lambda x: x**3 - 2 * x + 3, FI(0, 2))
    >>> print(format(y, ".3f"))
    [1.911, 1.912]
    >>> print(format(x[0], ".3f"))
    [0.816, 0.817]
    """
    if not isinstance(domain, Interval):
        raise TypeError

    intvl: type[T] = type(domain)
    ZERO = domain.operator.ZERO
    INFINITY = domain.operator.INFINITY

    if fprime is True:
        fprime = deriv(fun)

    if isinstance(xtol, (float, int)):
        xtol = domain.converter.fromfloat(float(xtol), strict=False)

    if not isinstance(xtol, intvl.endtype):
        raise TypeError

    if isinstance(ytol, (float, int)):
        ytol = domain.converter.fromfloat(float(ytol), strict=False)

    if not isinstance(ytol, intvl.endtype):
        raise TypeError

    if max_iter <= 0:
        raise ValueError

    branches = [domain]
    bound = inf = sup = INFINITY
    argmins: list[T] = []

    for i in range(max_iter):
        if not branches:
            break

        next_branches: list[T] = []

        for x in branches:
            y = fun(x)

            if y.inf > bound:
                continue

            x0 = x.mid()
            y0 = fun(intvl(x0))

            if fprime:
                c = fprime(x)

                if c.inf > ZERO and x.inf != domain.inf:
                    continue

                if c.sup < ZERO and x.sup != domain.sup:
                    continue

                y &= y0 + c * (x - x0)

                if y.inf > bound:
                    continue

            bound = min(bound, y0.sup)
            is_sufficient = x.diam() <= xtol and y.diam() <= abs(y.mid()) * ytol

            if is_sufficient or i == max_iter - 1:
                argmins.append(x)
                inf = min(inf, y.inf)
                sup = min(sup, y.sup)
                continue

            next_branches.extend((intvl(x.inf, x0), intvl(x0, x.sup)))

        branches = next_branches

    return intvl(inf, sup), argmins



from verry import FloatInterval as FI
from verry import function as vrf
from verry.linalg import FloatIntervalMatrix as FIM
from verry.optimize.rootfinding import allroot, allroot_scalar


def test_allroot():
    domain = FIM(inf=[-3, -3], sup=[3, 3])
    r = allroot(lambda x, y: (x**2 - y - 3, -x + y**2 - 3), domain, unique=True)
    assert len(r.unique) == 4
    root_min = min(r.unique, key=lambda x: x[0].inf)
    assert -2 in root_min[0] and 1 in root_min[1]


def test_allroot_scalar():
    domain = FI(-2, 3)
    r = allroot_scalar(lambda x: x**3 - 2 * x, domain, unique=True)
    assert len(r.unique) == 3
    root_max = max(r.unique, key=lambda x: x.sup)
    assert root_max.issuperset(vrf.sqrt(FI(2)))



from abc import ABC, abstractmethod
from collections.abc import Callable, Sequence
from typing import Any, ClassVar, Literal

from verry import function as vrf
from verry.integrate.utility import seriessolution
from verry.interval.interval import Interval
from verry.intervalseries import IntervalSeries, localcontext
from verry.linalg.intervalmatrix import IntervalMatrix


class Integrator[T: Interval](ABC):
    r"""Abstract base class for ODE integrators.

    Parameters
    ----------
    fun : Callable
        Right-hand side of the system. The calling signature is ``fun(t, *y)``.
    t0 : Interval
        Initial time.
    y0 : IntervalMatrix | Sequence[Interval]
        Initial state.
    t_bound : Interval
        Boundary time. Its infimum must be greater than the supremum of `t0`.

    Attributes
    ----------
    status : Literal["FAILURE", "RUNNING", "SUCCESS"]
        Current status of the integrator.
    t : Interval
        Current time.
    y : tuple[Interval, ...]
        Current state.
    t_bound : Interval
        Boundary time.
    t_prev : Interval | None
        Previous time. ``None`` if no steps were made yet.
    series : tuple[IntervalSeries, ...] | None
        Current interval series. ``None`` if no steps were made yet.
    order : int
        Degree of `series`.

    Warnings
    --------
    `fun` must be a :math:`C^\infty`-function on some open interval that contains `t0`
    and `t_bound`. Furthermore, `fun` must neither be a constant nor contain conditional
    branches (cf. :doc:`/userguide/pitfall`).

    Notes
    -----
    The implementation of :class:`Integrator` must satisfy the following conditions:

    * `t` and `t_prev` are always disjoint.
    * If `status` is ``"RUNNING"``, the graph of the solution intersects the Cartesian
      product of `t` and `y`.
    * If `status` is ``"RUNNING"`` and the last called method is :meth:`step`, the
      infimum and supremum of `t` are equal.
    * If `status` is ``"SUCCESS"``, `t` equals `t_bound`, and `y` contains the image of
      `t_bound` under the solution.

    Here :math:`\phi(t;t_0,y_0)` denotes a time-dependent flow of ODEs. Suppose that
    :math:`(\hat{t}_{\mathrm{prev}},\hat{y}_{\mathrm{prev}})` is an arbitrary element of
    the Cartesian product of :math:`[t_{\mathrm{prev}}]` and
    :math:`[y_{\mathrm{prev}}]`. Let :math:`d` be `order`, and
    :math:`q_{\mathrm{apx}}(s)` be a (:math:`d-1`)-th order Taylor polynomial of
    :math:`\phi(\hat{t}_{\mathrm{prev}}+s;\hat{t}_{\mathrm{prev}},
    \hat{y}_{\mathrm{prev}})`. Also, let :math:`[p(s)]=[p_{\mathrm{apx}}(s)]+[a]s^d` be
    `series`, assuming that the degree of :math:`[p_{\mathrm{apx}}(s)]` is smaller than
    :math:`d`. Then :math:`[p(s)]` must satisfy the following conditions:

    * :math:`q_{\mathrm{apx}}(s)\in[p_{\mathrm{apx}}(s)]` holds coefficient-wise.
    * :math:`\phi(\hat{t}_{\mathrm{prev}}+s;\hat{t}_{\mathrm{prev}},
      \hat{y}_{\mathrm{prev}})-q_{\mathrm{apx}}(s)\in [a]s^d` holds for all
      :math:`s\in(0,t-\hat{t}_{\mathrm{prev}})`.
    """

    order: int
    series: tuple[IntervalSeries[T], ...] | None
    status: Literal["FAILURE", "RUNNING", "SUCCESS"]
    t: T
    t_bound: T
    t_prev: T | None
    y: tuple[T, ...]

    @abstractmethod
    def __init__(
        self,
        fun: Callable,
        t0: T,
        y0: IntervalMatrix[T] | Sequence[T],
        t_bound: T,
    ):
        raise NotImplementedError

    @abstractmethod
    def step(self) -> tuple[Literal[True], None] | tuple[Literal[False], str]:
        """Perform one integration step.

        Returns
        -------
        r0 : bool
            `r0` is ``True`` if and only if `status` is not ``"FAILURE"``.
        r1 : str | None
            `r1` is ``None`` If `r0` is ``True``; otherwise, `r1` describes the reason
            for failure.
        """
        raise NotImplementedError

    @abstractmethod
    def update(self, t: T, y: IntervalMatrix[T] | Sequence[T]) -> None:
        """Receive a state, usually refined by :class:`Tracker`.

        ``self.t`` and ``self.y`` are updated to `t` and `y`.

        Parameters
        ----------
        t : Interval
        y : IntervalMatrix | Sequence[Interval]

        Raises
        ------
        ValueError
            If `status` is ``"FAILURE"``.
        """
        raise NotImplementedError


def eilo[T: Interval = Any](
    order: int = 15,
    rtol: Any = 1e-10,
    atol: Any = 1e-10,
    max_tries: int = 5,
    min_step: Any = 0.0,
    max_step: Any = float("inf"),
) -> type[Integrator[T]]:
    r"""Integrator based on Eijgenraam and Lohner's algorithm.

    See their publications [#Ei81]_\ [#Lo87]_\ [#Lo92]_ for more details on the theory.

    Parameters
    ----------
    order : int, default=15
        Degree of Taylor polynomial. `order` must be greater than or equal to 3.
    rtol : default=1e-10
        Relative tolerance.
    atol : default=1e-10
        Absolute tolerance.
    max_tries : int, default=5
        Maximum number of trials to verify the existence conditions of the solution.
    min_step : default=0.0
        Allowed minimum step size. If adaptive stepping produced a step size that
        is smaller than `min_step`, step size is set to `min_step`.
    max_step : default=inf
        Allowed maximum step size. If adaptive stepping produced a step size that
        is greater than `max_step`, step size is set to `max_step`.

    Returns
    -------
    type[Integrator]

    References
    ----------
    .. [#Ei81] P. Eijgenraam, *The solution of initial value problems using interval
        arithmetic: formulation and analysis of an algorithm*. Amsterdam, Nederland:
        CWI, 1981. [Online]. Available: https://ir.cwi.nl/pub/13004/
    .. [#Lo87] R. J. Lohner, "Enclosing the Solutions of Ordinary Initial and Boundary
        Value Problem," in *Computerarithmetic*, E. Kaucher, U. Kulisch, and Ch.
        Ullrich, Eds. Stuttgart, Germany: B. G. Teubner, 1987, pp. 225--286.
    .. [#Lo92] R. J. Lohner, "Computation of guaranteed enclosures for the solutions of
        ordinary initial and boundary value problems," in *Computational Ordinary
        Differential Equations*, J. R. Cash and I. Gladwell, Eds. Oxford, UK: Clarendon
        Press, 1992, pp. 425--435.
    """

    if order < 3:
        raise ValueError

    if max_tries < 1:
        raise ValueError

    class Integrator(
        _eilo,
        order=order,
        rtol=rtol,
        atol=atol,
        max_tries=max_tries,
        min_step=min_step,
        max_step=max_step,
    ):
        pass

    return Integrator


class _eilo[T: Interval](Integrator[T], ABC):
    order: ClassVar[int]
    _max_tries: ClassVar[int]
    _cls_rtol: ClassVar[Any]
    _cls_atol: ClassVar[Any]
    _cls_min_step: ClassVar[Any]
    _cls_max_step: ClassVar[Any]

    status: Literal["FAILURE", "RUNNING", "SUCCESS"]
    t: T
    y: tuple[T, ...]
    t_bound: T
    t_prev: T | None
    series: tuple[IntervalSeries[T], ...] | None
    _fun: Callable
    _rtol: Any
    _atol: Any
    _min_step: Any
    _max_step: Any

    def __init__(
        self, fun: Callable, t0: T, y0: IntervalMatrix[T] | Sequence[T], t_bound: T
    ):
        rtol = self._cls_rtol
        atol = self._cls_atol
        min_step = self._cls_min_step
        max_step = self._cls_max_step

        if not isinstance(t0, Interval):
            raise TypeError

        if t0.sup >= t_bound.inf:
            raise ValueError

        if isinstance(rtol, float):
            rtol = t0.converter.fromfloat(rtol, strict=False)

        if isinstance(atol, float):
            atol = t0.converter.fromfloat(atol, strict=False)

        if isinstance(min_step, float):
            min_step = t0.converter.fromfloat(min_step)

        if isinstance(max_step, float):
            max_step = t0.converter.fromfloat(max_step)

        if not t0.operator.ZERO <= min_step <= max_step:
            raise ValueError

        self.status = "RUNNING"
        self.t = t0
        self.y = tuple(y0)
        self.t_bound = t_bound
        self.t_prev = None
        self.series = None
        self._fun = fun
        self._rtol = rtol
        self._atol = atol
        self._min_step = min_step
        self._max_step = max_step

    def step(self) -> tuple[Literal[True], None] | tuple[Literal[False], str]:
        ZERO = self.t.operator.ZERO
        intvl = type(self.t)
        cadd = self.t.operator.cadd
        csub = self.t.operator.csub

        p0 = seriessolution(self._fun, self.t, self.y, self.order)
        tmp = ZERO

        for i in (self.order, self.order - 1, self.order - 2):
            tmp = max(tmp, vrf.pow(max(x.coeffs[i].mag() for x in p0), 1 / i))

        if tmp != ZERO:
            tol = self._atol + self._rtol * max(abs(x.mid()) for x in self.y)
            stepsize = vrf.pow(tol, 1 / self.order) / tmp
        elif self.t_prev is not None:
            stepsize = self.t.mid() - self.t_prev.mid()
        else:
            stepsize = csub(self.t_bound.sup, self.t.inf)

        stepsize = min(max(stepsize, self._min_step), self._max_step)

        if stepsize == ZERO:
            self.status = "FAILURE"
            return (False, "failed to determine a step size")

        t_next = intvl(cadd(self.t.sup, stepsize))
        is_verified = False

        for _ in range(self._max_tries):
            if t_next.sup >= self.t_bound.inf:
                self.status = "SUCCESS"
                t_next = self.t_bound

            h = intvl(0, csub(t_next.sup, self.t.inf))
            t1 = self.t + h
            a0 = [y0 + h * dy for y0, dy in zip(self.y, self._fun(t1, *self.y))]

            for __ in range(5):
                a1 = [y0 + h * dy for y0, dy in zip(self.y, self._fun(t1, *a0))]

                if all(y.issubset(x) for x, y in zip(a0, a1)):
                    is_verified = True
                    break

                EPSILON = intvl("0.1")
                a0 = [(1 + EPSILON) * x - EPSILON * x for x in a1]

            if is_verified:
                break

            if stepsize == self._min_step:
                self.status = "FAILURE"
                return (False, "failed to determine a step size")

            self.status = "RUNNING"
            stepsize = max(stepsize / 2, self._min_step)

            if stepsize == ZERO:
                self.status = "FAILURE"
                return (False, "failed to determine a step size")

            t_next = intvl(cadd(self.t.sup, stepsize))

        if not is_verified:
            self.status = "FAILURE"
            return (False, f"failed to verify within {self._max_tries} trials")

        p1 = seriessolution(self._fun, self.t, a1, self.order)

        for i in range(len(p0)):
            p0[i].coeffs[-1] = p1[i].coeffs[-1]

        self.t_prev = self.t
        self.t = t_next
        self.y = tuple(x.eval(self.t - self.t_prev) for x in p0)
        self.series = tuple(p0)
        return (True, None)

    def update(self, t: T, y: IntervalMatrix[T] | Sequence[T]) -> None:
        if self.status == "FAILURE":
            raise RuntimeError

        if t != self.t_bound:
            self.status = "RUNNING"

        self.t = t.copy()
        self.y = tuple(x.copy() for x in y)

    def __init_subclass__(
        cls,
        /,
        order: int,
        rtol: Any,
        atol: Any,
        max_tries: int,
        min_step: Any,
        max_step: Any,
        **kwargs,
    ):
        super().__init_subclass__(**kwargs)

        if order < 3:
            raise ValueError

        if max_tries < 1:
            raise ValueError

        cls.order = order
        cls._max_tries = max_tries
        cls._cls_rtol = rtol
        cls._cls_atol = atol
        cls._cls_min_step = min_step
        cls._cls_max_step = max_step


def kashi[T: Interval = Any](
    order: int = 15,
    rtol: Any = 1e-10,
    atol: Any = 1e-10,
    max_tries: int = 5,
    min_step: Any = 0.0,
    max_step: Any = float("inf"),
) -> type[Integrator[T]]:
    r"""Integrator based on Kashiwagi's algorithm.

    This is an implementation of [#Ka95]_.

    Parameters
    ----------
    order : int, default=15
        Degree of Taylor polynomial. `order` must be greater than or equal to 3.
    rtol : default=1e-10
        Relative tolerance.
    atol : default=1e-10
        Absolute tolerance.
    max_tries : int, default=5
        Maximum number of trials to verify the existence conditions of the solution.
    min_step : default=0.0
        Allowed minimum step size. If adaptive stepping produced a step size that
        is smaller than `min_step`, step size is set to `min_step`.
    max_step : default=inf
        Allowed maximum step size. If adaptive stepping produced a step size that
        is greater than `max_step`, step size is set to `max_step`.

    Returns
    -------
    type[Integrator]

    References
    ----------
    .. [#Ka95] M. Kashiwagi, "Power series arithmetic and its application to numerical
        validation," in *Proc. 1995 Symposium on Nonlinear Theory and its Applications
        (NOLTA '95)*, Las Vegas, NV, USA, Dec. 10--14, 1995, pp. 251--254.
    """

    if order < 3:
        raise ValueError

    if max_tries < 1:
        raise ValueError

    class Integrator(
        _kashi,
        order=order,
        rtol=rtol,
        atol=atol,
        max_tries=max_tries,
        min_step=min_step,
        max_step=max_step,
    ):
        pass

    return Integrator


class _kashi[T: Interval](Integrator[T], ABC):
    order: ClassVar[int]
    _max_tries: ClassVar[int]
    _cls_rtol: ClassVar[Any]
    _cls_atol: ClassVar[Any]
    _cls_min_step: ClassVar[Any]
    _cls_max_step: ClassVar[Any]

    status: Literal["FAILURE", "RUNNING", "SUCCESS"]
    t: T
    y: tuple[T, ...]
    t_bound: T
    t_prev: T | None
    series: tuple[IntervalSeries[T], ...] | None
    _fun: Callable
    _rtol: Any
    _atol: Any
    _min_step: Any
    _max_step: Any

    def __init__(
        self, fun: Callable, t0: T, y0: IntervalMatrix[T] | Sequence[T], t_bound: T
    ):
        rtol = self._cls_rtol
        atol = self._cls_atol
        min_step = self._cls_min_step
        max_step = self._cls_max_step

        if not isinstance(t0, Interval):
            raise TypeError

        if t0.sup >= t_bound.inf:
            raise ValueError

        if isinstance(rtol, float):
            rtol = t0.converter.fromfloat(rtol, strict=False)

        if isinstance(atol, float):
            atol = t0.converter.fromfloat(atol, strict=False)

        if isinstance(min_step, float):
            min_step = t0.converter.fromfloat(min_step)

        if isinstance(max_step, float):
            max_step = t0.converter.fromfloat(max_step)

        if not t0.operator.ZERO <= min_step <= max_step:
            raise ValueError

        self.status = "RUNNING"
        self.t = t0
        self.y = tuple(y0)
        self.t_bound = t_bound
        self.t_prev = None
        self.series = None
        self._fun = fun
        self._rtol = rtol
        self._atol = atol
        self._min_step = min_step
        self._max_step = max_step

    def step(self) -> tuple[Literal[True], None] | tuple[Literal[False], str]:
        ZERO = self.t.operator.ZERO
        intvl = type(self.t)
        cadd = self.t.operator.cadd
        csub = self.t.operator.csub

        p0 = seriessolution(self._fun, self.t, self.y, self.order)
        tmp = ZERO

        for i in (self.order, self.order - 1, self.order - 2):
            tmp = max(tmp, vrf.pow(max(x.coeffs[i].mag() for x in p0), 1 / i))

        if tmp != ZERO:
            tol = self._atol + self._rtol * max(abs(x.mid()) for x in self.y)
            stepsize = vrf.pow(tol, 1 / self.order) / tmp
        elif self.t_prev is not None:
            stepsize = self.t.mid() - self.t_prev.mid()
        else:
            stepsize = csub(self.t_bound.sup, self.t.inf)

        stepsize = min(max(stepsize, self._min_step), self._max_step)

        if stepsize == ZERO:
            self.status = "FAILURE"
            return (False, "failed to determine a step size")

        t_next = intvl(cadd(self.t.sup, stepsize))

        if t_next.sup >= self.t_bound.inf:
            t_next = self.t_bound

        with localcontext() as ctx:
            ctx.rounding = "TYPE2"
            ctx.deg = self.order
            ctx.domain = intvl(0, csub(t_next.sup, self.t.inf))

            t = IntervalSeries([self.t, 1], intvl=intvl)
            p2 = [y0 + dy.integrate() for y0, dy in zip(self.y, self._fun(t, *p0))]

            for i in range(len(p2)):
                ctx.round(p2[i])

            rad = max((x.coeffs[-1] - y.coeffs[-1]).mag() for x, y in zip(p0, p2))
            p1 = [x.copy() for x in p0]

            for i in range(len(p1)):
                p1[i].coeffs[-1] += rad * intvl(-2, 2)

            p2 = [y0 + dy.integrate() for y0, dy in zip(self.y, self._fun(t, *p1))]

            for i in range(len(p2)):
                ctx.round(p2[i])

            tmp = [x.eval(stepsize) for x in p2]
            e0 = self._atol + self._rtol * max(abs(x.mid()) for x in tmp)
            e1 = max(x.diam() for x in tmp)

            if e1 > ZERO:
                stepsize *= vrf.pow(e0 / e1, 1 / self.order)
                stepsize = min(max(stepsize, self._min_step), self._max_step)

                if stepsize == ZERO:
                    self.status = "FAILURE"
                    return (False, "failed to determine a step size")

            t_next = intvl(cadd(self.t.sup, stepsize))

            for _ in range(self._max_tries):
                if t_next.sup >= self.t_bound.inf:
                    self.status = "SUCCESS"
                    t_next = self.t_bound

                ctx.domain = intvl(0, csub(t_next.sup, self.t.inf))
                t = IntervalSeries([self.t, 1], intvl=intvl)
                p2 = [y0 + dy.integrate() for y0, dy in zip(self.y, self._fun(t, *p1))]

                for i in range(len(p2)):
                    ctx.round(p2[i])

                if all(y.coeffs[-1].issubset(x.coeffs[-1]) for x, y in zip(p1, p2)):
                    self.t_prev = self.t
                    self.t = t_next
                    self.y = tuple(x.eval(self.t - self.t_prev) for x in p2)
                    self.series = tuple(p2)
                    return (True, None)

                if stepsize == self._min_step:
                    self.status = "FAILURE"
                    return (False, "failed to determine a step size")

                self.status = "RUNNING"
                stepsize = max(stepsize / 2, self._min_step)

                if stepsize == ZERO:
                    self.status = "FAILURE"
                    return (False, "failed to determine a step size")

                t_next = intvl(cadd(self.t.sup, stepsize))

        self.status = "FAILURE"
        return (False, f"failed to verify within {self._max_tries} trials")

    def update(self, t: T, y: IntervalMatrix[T] | Sequence[T]) -> None:
        if self.status == "FAILURE":
            raise RuntimeError

        if t != self.t_bound:
            self.status = "RUNNING"

        self.t = t.copy()
        self.y = tuple(x.copy() for x in y)

    def __init_subclass__(
        cls,
        /,
        order: int,
        rtol: Any,
        atol: Any,
        max_tries: int,
        min_step: Any,
        max_step: Any,
        **kwargs,
    ):
        super().__init_subclass__(**kwargs)
        cls.order = order
        cls._max_tries = max_tries
        cls._cls_rtol = rtol
        cls._cls_atol = atol
        cls._cls_min_step = min_step
        cls._cls_max_step = max_step



import functools
from collections.abc import Callable
from typing import Any

from verry.autodiff._dual import Dual, Vector


def deriv[T, **P](fun: Callable[P, T]) -> Callable[P, T]:
    """Return a function that evaluates the derivative of the univariate scalar-valued
    function.

    Parameters
    ----------
    fun : Callable
        Diffarentiated function.

    Returns
    -------
    Callable
        Derivative of `fun`.

    Warnings
    --------
    `fun` must neither be a constant nor contain conditional branches
    (cf. :doc:`/userguide/pitfall`).

    Examples
    --------
    This example computes a value of the derivative.

    >>> from verry import FloatInterval as FI
    >>> from verry import function as vrf
    >>> f = lambda x: x**2 + vrf.sqrt(x + 3)
    >>> df = deriv(f)
    >>> c0 = df(1.2)
    >>> print(format(c0, ".6g"))
    2.64398

    To obtain the value that accounts for rounding errors, pass an interval to `df`.

    >>> c1 = df(FI("1.2"))
    >>> print(format(c1, ".6g"))
    [2.64397, 2.64398]

    Also, the second-order derivative can be obtained in the same way.

    >>> ddf = deriv(df)
    >>> print(format(ddf(FI("1.2")), ".6g"))
    [1.97095, 1.97096]
    """

    def result(*args, **kwargs):
        tmp: Any = fun(*Dual.variable(*args), **kwargs)  # type: ignore
        return tmp.imag[0]

    return result


def grad[T, **P](fun: Callable[P, T]) -> Callable[P, tuple[T, ...]]:
    """Return a function that evaluates the gradient of the multivariate scalar-valued
    function.

    Parameters
    ----------
    fun : Callable
        Diffarentiated function.

    Returns
    -------
    Callable
        Gradient of `fun`.

    Warnings
    --------
    `fun` must neither be a constant nor contain conditional branches
    (cf. :doc:`/userguide/pitfall`).

    Examples
    --------
    This example computes a value of the gradient.

    >>> from verry import FloatInterval as FI
    >>> from verry import function as vrf
    >>> f = lambda x, y: vrf.sqrt(x * y + 3)
    >>> df = grad(f)
    >>> c0 = df(0.5, 1.0)
    >>> print(format(c0[0], ".6g"), format(c0[1], ".6g"))
    0.267261 0.133631

    To obtain the value that accounts for rounding errors, pass intervals to `df`.

    >>> c1 = df(FI("0.5"), FI("1"))
    >>> print(format(c1[0], ".6g"))
    [0.267261, 0.267262]
    """

    def result(*args, **kwargs):
        tmp: Any = fun(*Dual.variable(*args), **kwargs)  # type: ignore
        return tuple(tmp.imag)

    return result


def jacobian[T: tuple, **P](fun: Callable[P, T]) -> Callable[P, tuple[T, ...]]:
    """Return a function that evaluates the Jacobian matrix of the multivariate
    vector-valued function.

    Parameters
    ----------
    fun : Callable
        Diffarentiated function.

    Returns
    -------
    Callable
        Fr\u00e9chet derivative of `fun`.

    Warnings
    --------
    `fun` must neither be a constant nor contain conditional branches
    (cf. :doc:`/userguide/pitfall`).
    """

    def result(*args, **kwargs):
        tmp: Any = fun(*Dual.variable(*args), **kwargs)  # type: ignore
        return tuple(tuple(x.imag) for x in tmp)

    return result


def _defderiv[**P](
    fun: Callable[P, Any], deriv: Callable[P, Any], *, argnum: int = 0
) -> None:
    if "_verry_is_primitive" not in fun.__dict__:
        raise ValueError

    fun.__dict__["_verry_derivs"][argnum] = deriv


def _primitive[T, **P](fun: Callable[P, T]) -> Callable[P, T]:
    derivs: dict[int, Callable] = {}

    @functools.wraps(fun)
    def wrapper(*args, **kwargs):
        max_level = max((x.level for x in args if isinstance(x, Dual)), default=0)

        if max_level == 0:
            return fun(*args, **kwargs)

        real: list = []
        imag: list[tuple[int, Vector]] = []

        for i, x in enumerate(args):
            if isinstance(x, Dual) and x.level == max_level:
                real.append(x.real)
                imag.append((i, x.imag))
            else:
                real.append(x)

        tmp = [derivs[i](*real, **kwargs) * x for i, x in imag]
        return Dual(wrapper(*real, **kwargs), sum(tmp[1:], start=tmp[0]))

    wrapper.__dict__["_verry_is_primitive"] = True
    wrapper.__dict__["_verry_derivs"] = derivs
    return wrapper  # type: ignore



import re
from typing import Literal, Self, TypedDict, Unpack

_PATTERN = re.compile(
    r"(?:(?P<fill>[\s\S])?(?P<align>[<>=^]))?"
    r"(?P<sign>[-+ ])?"
    r"(?P<z>z)?"
    r"(?P<alt>#)?"
    r"(?P<zfill>0)?"
    r"(?P<width>\d+)?"
    r"(?P<grouping>[_,])?"
    r"(?:\.(?P<prec>\d+))?"
    r"(?P<type>[eEfFgGn%])?"
)


class _FormatSpecDict(TypedDict, total=False):
    fill: str
    align: Literal["<", ">", "=", "^"] | None
    sign: Literal["+", "-", "\u0020"]
    z: bool
    alt: bool
    zfill: bool
    width: int | None
    grouping: Literal["_", ","] | None
    prec: int | None
    type: Literal["e", "E", "f", "F", "g", "G", "n", "%"] | None


class FormatSpec:
    r"""Utility for managing format specifications.

    See `Python's documentation
    <https://docs.python.org/3/library/string.html#formatspec>`__ for the meaning of
    each parameter.

    Parameters
    ----------
    format_spec : str, default=""
    fill : str, default="\\u0020"
    align : Literal["<", ">", "=", "^"] | None, default=None
    sign : Literal["+", "-", "\\u0020"], default="-"
    z : bool, default=False
    alt : bool, default=False
    zfill : bool, default=False
    width : int | None, default=None
    grouping : Literal["_", ","] | None, default=None
    prec : int | None, default=None
    type : Literal["e", "E", "f", "F", "g", "G", "n", "%"] | None, default=None

    Attributes
    ----------
    fill : str
    align : Literal["<", ">", "=", "^"] | None
    sign : Literal["+", "-", "\\u0020"]
    z : bool
    alt : bool
    zfill : bool
    width : int | None
    grouping : Literal["_", ","] | None
    prec : int | None
    type : Literal["e", "E", "f", "F", "g", "G", "n", "%"] | None

    Examples
    --------
    >>> x = FormatSpec(".3f", prec=4)
    >>> x
    FormatSpec('.4f')
    >>> str(x)
    '.4f'
    >>> x.prec
    4
    >>> x.type
    'f'
    """

    __slots__ = (
        "fill",
        "align",
        "sign",
        "z",
        "alt",
        "zfill",
        "width",
        "grouping",
        "prec",
        "type",
    )

    fill: str
    align: Literal["<", ">", "=", "^"] | None
    sign: Literal["+", "-", "\u0020"]
    z: bool
    alt: bool
    zfill: bool
    width: int | None
    grouping: Literal["_", ","] | None
    prec: int | None
    type: Literal["e", "E", "f", "F", "g", "G", "n", "%"] | None

    def __init__(self, format_spec: str = "", **kwargs: Unpack[_FormatSpecDict]):
        self.fill = "\u0020"
        self.align = None
        self.sign = "-"
        self.z = False
        self.alt = False
        self.zfill = False
        self.width = None
        self.grouping = None
        self.prec = None
        self.type = None

        if format_spec:
            if (match := _PATTERN.fullmatch(format_spec)) is None:
                raise ValueError

            if (fill := match.group("fill")) is not None:
                self.fill = fill

            if (align := match.group("align")) is not None:
                self.align = align  # type: ignore

            if (sign := match.group("sign")) is not None:
                self.sign = sign  # type: ignore

            self.z = match.group("z") is not None
            self.alt = match.group("alt") is not None
            self.zfill = match.group("zfill") is not None

            if (width := match.group("width")) is not None:
                self.width = int(width)

            self.grouping = match.group("grouping")  # type: ignore

            if (prec := match.group("prec")) is not None:
                self.prec = int(prec)

            self.type = match.group("type")  # type: ignore

        for key, value in kwargs.items():
            setattr(self, key, value)

    def copy(self) -> Self:
        result = self.__class__()
        result.fill = self.fill
        result.align = self.align
        result.sign = self.sign
        result.z = self.z
        result.alt = self.alt
        result.zfill = self.zfill
        result.width = self.width
        result.grouping = self.grouping
        result.prec = self.prec
        result.type = self.type
        return result

    def format(self, value: object) -> str:
        """Shorthand for ``format(value, str(self))``."""
        return format(value, self.__str__())

    def replace(self, **changes: Unpack[_FormatSpecDict]) -> Self:
        """Create a new :class:`FormatSpec`, replacing fields with values from
        `changes`."""
        result = self.copy()

        for key, value in changes.items():
            setattr(result, key, value)

        return result

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.__str__()!r})"

    def __str__(self) -> str:
        result = ""

        if self.align is not None:
            if self.fill != "\u0020":
                result += self.fill

            result += self.align

        if self.sign != "-":
            result += self.sign

        if self.z:
            result += "z"

        if self.alt:
            result += "#"

        if self.zfill:
            result += "0"

        if self.width is not None:
            result += str(self.width)

        if self.grouping is not None:
            result += self.grouping

        if self.prec is not None:
            result += f".{self.prec}"

        if self.type is not None:
            result += self.type

        return result

    def __eq__(self, other: Self) -> bool:
        if type(other) is not type(self):
            return NotImplemented

        return (
            other.fill == self.fill
            and other.align == self.align
            and other.sign == self.sign
            and other.z == self.z
            and other.alt == self.alt
            and other.zfill == self.zfill
            and other.width == self.width
            and other.grouping == self.grouping
            and other.prec == self.prec
            and other.type == self.type
        )

    def __bool__(self) -> bool:
        return (
            self.fill != "\u0020"
            or self.align is not None
            or self.sign != "-"
            or self.z
            or self.alt
            or self.zfill
            or self.width is not None
            or self.grouping is not None
            or self.prec is not None
            or self.type is not None
        )

    def __copy__(self) -> Self:
        return self.copy()

    def __replace__(self, **changes: Unpack[_FormatSpecDict]) -> Self:
        return self.replace(**changes)



from verry import function as vrf
from verry.integrate.integrator import eilo, kashi
from verry.integrate.solver import C0Solver, C1Solver
from verry.interval.floatinterval import FloatInterval as FI


def test_a2detest():
    solver = C0Solver(eilo)
    r = solver.solve(lambda _, x: (-(x**3) / 2,), FI(0), [FI(1)], FI(8))
    assert r.status == "SUCCESS"
    assert 1.0 / 3.0 in r.content.y[0]


def test_a4detest():
    a = FI("0.0125")
    t_bound = -4 * vrf.log(FI(9) / FI(19))
    solver = C1Solver(kashi)
    r = solver.solve(lambda _, x: (a * x * (20 - x),), FI(0), [FI(1)], t_bound)
    assert r.status == "SUCCESS"
    assert 2 in r.content.y[0]


def test_explosion():
    solver = C0Solver(eilo(min_step=1e-4))
    r = solver.solve(lambda _, x: (x**2,), FI(0), [FI(1)], FI(2))
    assert r.status == "FAILURE"



from collections.abc import Sequence
from typing import Any, Self, overload

import numpy as np
import numpy.typing as npt

from verry.interval.floatinterval import FloatInterval
from verry.linalg.intervalmatrix import (
    IntervalMatrix,
    approx_inv,
    approx_qr,
    approx_solve,
)


class FloatIntervalMatrix(IntervalMatrix[FloatInterval, float]):
    __slots__ = ()

    @overload
    def __init__(
        self,
        inf: IntervalMatrix[FloatInterval]
        | npt.NDArray
        | Sequence[FloatInterval | float | int | str]
        | Sequence[npt.NDArray | Sequence[FloatInterval | float | int | str]],
        /,
        *,
        intvl: Any = None,
    ): ...

    @overload
    def __init__(
        self,
        inf: npt.NDArray
        | Sequence[float | int | str]
        | Sequence[npt.NDArray | Sequence[float | int | str]],
        sup: npt.NDArray
        | Sequence[float | int | str]
        | Sequence[npt.NDArray | Sequence[float | int | str]],
        *,
        intvl: Any = None,
    ): ...

    def __init__(self, inf, sup=None, *, intvl=None, **kwargs):
        super().__init__(inf, sup, intvl=FloatInterval, **kwargs)

    @classmethod
    def empty(
        cls, shape: int | tuple[int] | tuple[int, int], *, intvl: Any = None
    ) -> Self:
        return super().empty(shape, intvl=FloatInterval)

    @classmethod
    def eye(cls, n: int, m: int | None = None, *, intvl: Any = None) -> Self:
        return super().eye(n, m, intvl=FloatInterval)

    @classmethod
    def ones(
        cls, shape: int | tuple[int] | tuple[int, int], *, intvl: Any = None
    ) -> Self:
        return super().ones(shape, intvl=FloatInterval)

    @classmethod
    def zeros(
        cls, shape: int | tuple[int] | tuple[int, int], *, intvl: Any = None
    ) -> Self:
        return super().zeros(shape, intvl=FloatInterval)

    @classmethod
    def _emptyarray(cls, shape):
        return np.empty(shape, np.float64)

    def _verry_overload_(self, fun, *args, **kwargs):
        if fun is approx_inv:
            return np.linalg.inv(self.mid())

        if fun is approx_qr:
            return tuple(np.linalg.qr(self.mid()))

        if fun is approx_solve:
            a = args[0].mid() if isinstance(args[0], IntervalMatrix) else args[0]
            b = args[1].mid() if isinstance(args[1], IntervalMatrix) else args[1]
            return np.linalg.solve(a, b)

        return super()._verry_overload_(fun, *args, **kwargs)



"""
#####################################################
Optimization and root finding (:mod:`verry.optimize`)
#####################################################

.. currentmodule:: verry.optimize

This module provides solvers for root finding.

Global optimization
===================

.. autosummary::
    :toctree: generated/

    branchbound
    branchbound_scalar

Root finding
============

.. autosummary::
    :toctree: generated/

    allroot
    allroot_scalar
    krawczyk
    krawczyk_scalar

Miscellaneous
=============

.. autosummary::
    :toctree: generated/

    AllRootResult
    AllRootScalarResult

"""

from .optimize import branchbound, branchbound_scalar
from .rootfinding import (
    AllRootResult,
    AllRootScalarResult,
    allroot,
    allroot_scalar,
    krawczyk,
    krawczyk_scalar,
)

__all__ = [
    "branchbound",
    "branchbound_scalar",
    "AllRootResult",
    "AllRootScalarResult",
    "allroot",
    "allroot_scalar",
    "krawczyk",
    "krawczyk_scalar",
]



"""
#######################################
Affine arithmetic (:mod:`verry.affine`)
#######################################

.. currentmodule:: verry.affine

This module provides affine arithmetic.

Affine form
===========

.. autosummary::
    :toctree: generated/

    AffineForm
    summarize
    summarized

Context
=======

.. autosummary::
    :toctree: generated/

    Context
    getcontext
    localcontext
    setcontext

"""

import contextlib
import contextvars
from collections.abc import Sequence
from typing import Any, Literal, Never, Self

from verry import function as vrf
from verry.interval.interval import Interval
from verry.typing import ComparableScalar, Scalar


class Context:
    """Create a new context.

    Context can be regarded as a collection of noise symbols. Each instance of
    :class:`AffineForm` belongs to one context, and affine forms belonging to
    different contexts are considered independent.

    Parameters
    ----------
    rounding : Literal["BRUTE", "FAST"], default="BRUTE"
        Rounding mode. If `rounding` is ``"FAST"``, no new noise symbols are appended
        by addition, subtraction, or constant multiplication. Thus, the operations are
        relatively fast. Instead, the radius of the resulting interval may be increased.
    """

    __slots__ = ("_rounding", "_count")
    _rounding: Literal["BRUTE", "FAST"]
    _count: int

    def __init__(self, rounding: Literal["BRUTE", "FAST"] = "BRUTE"):
        self._rounding = rounding
        self._count = 0

    @property
    def rounding(self) -> Literal["BRUTE", "FAST"]:
        return self._rounding

    def copy(self) -> Self:
        return self.__class__(self._rounding)

    def create_noisesymbol(self) -> int:
        result = self._count
        self._count += 1
        return result

    def __str__(self):
        return f"{type(self).__name__}({self._rounding!r})"

    def __copy__(self) -> Self:
        return self.copy()


_var: contextvars.ContextVar[Context] = contextvars.ContextVar("affine")


def getcontext() -> Context:
    """Return the current context for the active thread."""
    if context := _var.get(None):
        return context

    context = Context()
    _var.set(context)
    return context


def setcontext(ctx: Context) -> None:
    """Set the current context for the active thread to `ctx`."""
    _var.set(ctx)


@contextlib.contextmanager
def localcontext(
    ctx: Context | None = None, *, rounding: Literal["FAST", "BRUTE"] | None = None
):
    """Return a context manager that will set the current context for the active thread
    to a copy of `ctx` on entry to the with-statement and restore the previous context
    when exiting the with-statement."""
    if ctx is None:
        ctx = getcontext()

    if rounding is None:
        rounding = ctx._rounding

    ctx = Context(rounding)
    token = _var.set(ctx)

    try:
        yield ctx
    finally:
        _var.reset(token)


class AffineForm[T1: Interval, T2: ComparableScalar = Any](Scalar):
    """Affine form.

    Parameters
    ----------
    value : Interval
        Interval transformed into a new affine form that is independent of all existing
        affine forms.

    Examples
    --------
    >>> from verry import FloatInterval
    >>> x0 = FloatInterval("-0.1", "0.1")
    >>> y0 = (x0 + 1)**2 - 2 * x0
    >>> print(format(y0, ".6f"))
    [0.609999, 1.410001]
    >>> x1 = AffineForm(x0)
    >>> y1 = (x1 + 1)**2 - 2 * x1
    >>> print(format(y1.range(), ".6f"))
    [0.989999, 1.010001]
    """

    __slots__ = ("_mid", "_coeffs", "_excess", "_intvl", "_context")
    _mid: T2
    _coeffs: dict[int, T2]
    _excess: T2
    _intvl: type[T1]
    _context: Context

    def __init__(self, value: T1, **kwargs: Never):
        if kwargs.get("_skipinit", False):
            return

        if not isinstance(value, Interval):
            raise TypeError

        self._intvl = type(value)
        self._coeffs = {}
        self._excess = value.operator.ZERO
        self._context = getcontext()

        key = self._context.create_noisesymbol()
        self._mid = value.mid()
        self._coeffs[key] = value.rad()

    @property
    def context(self) -> Context:
        return self._context

    @property
    def interval(self) -> type[T1]:
        return self._intvl

    @classmethod
    def zero(cls, intvl: type[T1], ctx: Context | None = None) -> Self:
        """Return an affine form consisting only of the constant term 0.

        Parameters
        ----------
        intvl : type[Interval]
            Interval type.
        ctx : Context, optional
            Context to which the affine form belongs. If no context is specified, the
            current context is used.
        """
        if not issubclass(intvl, Interval):
            raise TypeError

        if ctx is None:
            ctx = getcontext()

        result = cls(None, _skipinit=True)  # type: ignore
        result._mid = intvl.operator.ZERO
        result._coeffs = {}
        result._excess = intvl.operator.ZERO
        result._intvl = intvl
        result._context = ctx
        return result

    def copy(self) -> Self:
        """Return a shallow copy of the affine form."""
        result = self.zero(self._intvl, self._context)
        result._mid = self._mid
        result._coeffs = self._coeffs.copy()
        result._excess = self._excess
        return result

    def mid(self) -> T2:
        return self._mid

    def rad(self) -> T2:
        """Return an upper bound of the radius."""
        result = self._excess

        for value in self._coeffs.values():
            result = self._intvl.operator.cadd(result, abs(value))

        return result

    def range(self) -> T1:
        """Return the range of the affine form.

        Example
        -------
        >>> from verry import FloatInterval
        >>> x = AffineForm(FloatInterval("0.1"))
        >>> x.range() == x.mid() + x.rad() * x.interval(-1, 1)
        True
        """
        rad = self.rad()
        return self._mid + self._intvl(-rad, rad)

    def reciprocal(self) -> Self:
        """Return the reciprocal of the affine form.

        Raises
        ------
        ZeroDivisionError
            If the range contains zero.
        """
        ZERO = self._intvl.operator.ZERO
        cadd = self._intvl.operator.cadd
        cmul = self._intvl.operator.cmul

        if ZERO in (range := self.range()):
            raise ZeroDivisionError

        context = getcontext()
        self._ensurecontext(context)

        result = self.zero(self._intvl, context)
        tmp = self._intvl(range.inf)
        inv = -1 / (tmp * range.sup)
        am = (tmp + range.sup) / 2
        gm = vrf.sqrt(tmp * range.sup)

        if range.inf > ZERO:
            tmp = inv * (self._mid - (am + gm))
            result._mid = tmp.mid()
            error = cadd((inv * (am - gm)).mag(), tmp.rad())
        else:
            tmp = inv * (self._mid - (am - gm))
            result._mid = tmp.mid()
            error = cadd((inv * (am + gm)).mag(), tmp.rad())

        for key, coeff in self._coeffs.items():
            tmp = inv * coeff
            result._coeffs[key] = tmp.mid()
            error = cadd(error, tmp.rad())

        if context._rounding == "FAST":
            error = cadd(error, cmul(self._excess, inv.mag()))
            result._excess = ZERO

        key = context.create_noisesymbol()
        result._coeffs[key] = error
        return result

    def _ensurecontext(self, ctx: Context) -> None:
        if self._context is ctx:
            return

        coeffs = {}

        for value in self._coeffs.values():
            key = ctx.create_noisesymbol()
            coeffs[key] = value

        if self._context._rounding == "FAST" and ctx._rounding == "BRUTE":
            key = ctx.create_noisesymbol()
            coeffs[key] = self._excess
            self._excess = self._intvl.operator.ZERO

        self._coeffs.clear()
        self._coeffs = coeffs
        self._context = ctx

    def _verry_overload_(self, fun, *args, **kwargs):
        match fun:
            case vrf.e:
                return self.__class__(vrf.e(self._intvl()))

            case vrf.ln2:
                return self.__class__(vrf.ln2(self._intvl()))

            case vrf.pi:
                return self.__class__(vrf.pi(self._intvl()))

        return NotImplemented

    def __eq__(self, other) -> bool:
        if type(other) is not type(self):
            return NotImplemented

        return (
            other._intvl is self._intvl
            and other._mid == self._mid
            and other._coeffs == self._coeffs
            and other._excess == self._excess
            and other._context is self._context
        )

    def __len__(self) -> int:
        """Return the number of noise symbols."""
        return len(self._coeffs)

    def __add__(self, rhs: Self | T1 | T2 | float | int) -> Self:
        ZERO = self._intvl.operator.ZERO
        cadd = self._intvl.operator.cadd

        context = getcontext()
        self._ensurecontext(context)

        match rhs:
            case self._intvl.endtype() | float() | int():
                result = self.zero(self._intvl, context)
                tmp = self._intvl(self._mid) + rhs
                result._mid = tmp.mid()
                result._coeffs = self._coeffs.copy()
                error = tmp.rad()

                match context._rounding:
                    case "BRUTE":
                        key = context.create_noisesymbol()
                        result._coeffs[key] = error
                        return result

                    case "FAST":
                        error = cadd(error, self._excess)
                        result._excess = error
                        return result

            case self.__class__():
                rhs._ensurecontext(context)
                result = self.zero(self._intvl, context)
                tmp = self._intvl(self._mid) + rhs._mid
                result._mid = tmp.mid()
                error = tmp.rad()

                for key in set(self._coeffs) | set(rhs._coeffs):
                    x = self._intvl(self._coeffs.get(key, ZERO))
                    y = self._intvl(rhs._coeffs.get(key, ZERO))
                    tmp = x + y
                    result._coeffs[key] = tmp.mid()
                    error = cadd(error, tmp.rad())

                match context._rounding:
                    case "BRUTE":
                        key = context.create_noisesymbol()
                        result._coeffs[key] = error
                        return result

                    case "FAST":
                        error = cadd(error, cadd(self._excess, rhs._excess))
                        result._excess = error
                        return result

            case self._intvl():
                return self.__add__(self.__class__(rhs))

            case _:
                return NotImplemented

    def __sub__(self, rhs: Self | T1 | T2 | float | int) -> Self:
        ZERO = self._intvl.operator.ZERO
        cadd = self._intvl.operator.cadd

        context = getcontext()
        self._ensurecontext(context)

        match rhs:
            case self._intvl.endtype() | float() | int():
                result = self.zero(self._intvl, context)
                tmp = self._intvl(self._mid) - rhs
                result._mid = tmp.mid()
                result._coeffs = self._coeffs.copy()
                error = tmp.rad()

                match context._rounding:
                    case "BRUTE":
                        key = context.create_noisesymbol()
                        result._coeffs[key] = error
                        return result

                    case "FAST":
                        error = cadd(error, self._excess)
                        result._excess = error
                        return result

            case self.__class__():
                rhs._ensurecontext(context)
                result = self.zero(self._intvl, context)
                tmp = self._intvl(self._mid) - rhs._mid
                result._mid = tmp.mid()
                error = tmp.rad()

                for key in set(self._coeffs) | set(rhs._coeffs):
                    x = self._intvl(self._coeffs.get(key, ZERO))
                    y = self._intvl(rhs._coeffs.get(key, ZERO))
                    tmp = x - y
                    result._coeffs[key] = tmp.mid()
                    error = cadd(error, tmp.rad())

                match context._rounding:
                    case "BRUTE":
                        key = context.create_noisesymbol()
                        result._coeffs[key] = error
                        return result

                    case "FAST":
                        error = cadd(error, cadd(self._excess, rhs._excess))
                        result._excess = error
                        return result

            case self._intvl():
                return self.__sub__(self.__class__(rhs))

            case _:
                return NotImplemented

    def __mul__(self, rhs: Self | T1 | T2 | float | int) -> Self:
        ZERO = self._intvl.operator.ZERO
        cadd = self._intvl.operator.cadd
        cmul = self._intvl.operator.cmul

        context = getcontext()
        self._ensurecontext(context)

        match rhs:
            case self._intvl.endtype() | float() | int():
                rhs = self._intvl(rhs)
                result = self.zero(self._intvl, context)
                tmp = self._mid * rhs
                result._mid = tmp.mid()
                error = tmp.rad()

                for key, coeff in self._coeffs.items():
                    tmp = coeff * rhs
                    result._coeffs[key] = tmp.mid()
                    error = cadd(error, tmp.rad())

                match context._rounding:
                    case "BRUTE":
                        key = context.create_noisesymbol()
                        result._coeffs[key] = error
                        return result

                    case "FAST":
                        error = cadd(error, cmul(self._excess, rhs.mag()))
                        result._excess = error
                        return result

            case self.__class__():
                rhs._ensurecontext(context)
                result = self.zero(self._intvl, context)
                tmp = self._intvl(self._mid) * rhs._mid
                result._mid = tmp.mid()
                error = cadd(cmul(self.rad(), rhs.rad()), tmp.rad())

                for key in set(self._coeffs) | set(rhs._coeffs):
                    x = self._intvl(self._coeffs.get(key, ZERO))
                    y = self._intvl(rhs._coeffs.get(key, ZERO))
                    tmp = x * rhs._mid + self._mid * y
                    result._coeffs[key] = tmp.mid()
                    error = cadd(error, tmp.rad())

                if context._rounding == "FAST":
                    error = cadd(error, cmul(abs(self._mid), rhs._excess))
                    error = cadd(error, cmul(self._excess, abs(rhs._mid)))
                    result._excess = ZERO

                key = context.create_noisesymbol()
                result._coeffs[key] = error
                return result

            case self._intvl():
                return self.__mul__(self.__class__(rhs))

            case _:
                return NotImplemented

    def __truediv__(self, rhs: Self | T1 | T2 | float | int) -> Self:
        match rhs:
            case self._intvl.endtype() | float() | int():
                return self.__mul__(1 / self._intvl(rhs))

            case self.__class__():
                return self.__mul__(rhs.reciprocal())

            case self._intvl():
                return self.__mul__(self.__class__(rhs).reciprocal())

            case _:
                return NotImplemented

    def __pow__(self, rhs: int) -> Self:
        if not isinstance(rhs, int):
            return NotImplemented

        if rhs < 0:
            return self.__pow__(-rhs).reciprocal()

        result = self.__class__(self._intvl(1))
        tmp = self.copy()

        while rhs != 0:
            if rhs % 2 != 0:
                result *= tmp

            rhs //= 2
            tmp *= tmp

        return result

    def __radd__(self, lhs: Self | T1 | T2 | float | int) -> Self:
        return self.__add__(lhs)

    def __rsub__(self, lhs: Self | T1 | T2 | float | int) -> Self:
        return self.__neg__().__add__(lhs)

    def __rmul__(self, lhs: Self | T1 | T2 | float | int) -> Self:
        return self.__mul__(lhs)

    def __rtruediv__(self, lhs: Self | T1 | T2 | float | int) -> Self:
        return self.reciprocal().__mul__(lhs)

    def __neg__(self) -> Self:
        result = self.zero(self._intvl, self._context)
        result._mid = -self._mid
        result._coeffs = {key: -value for key, value in self._coeffs.items()}
        result._excess = self._excess
        return result

    def __pos__(self) -> Self:
        return self.copy()

    def __copy__(self) -> Self:
        return self.copy()


def summarized[T: AffineForm](vars: Sequence[T], n: int, m: int = 0) -> tuple[T, ...]:
    """Reduce the number of noise symbols while keeping correlation between `vars`.

    Parameters
    ----------
    vars : Sequence[AffineForm]
        Affine forms to which the summarization is applied.
    n : int
        Number of noise symbols after summarization. `n` must be greater than or equal
        to ``len(vars)``.
    m : int, default=0
        Threshold of summarization. If the number of noise symbols is less than `m`,
        summarization is not applied.

    Returns
    -------
    tuple[AffineForm, ...]

    Notes
    -----
    This is an implementation of [#Ka12]_.

    References
    ----------
    .. [#Ka12] M. Kashiwagi, "An algorithm to reduce the number of dummy variables in
        affine arithmetic," in *Proc. 15th GAMM-IMACS International Symposium on
        Scientific Computing Computer Arithmetic and Verified Numerical Computations
        (SCAN 2012)*, Novosibirsk, Russia, Sep. 23--29, 2012, pp. 70--71. [Online].
        Available: http://conf.nsc.ru/scan2012/scan2012_27
    """
    result = tuple(var.copy() for var in vars)
    summarize(result, n, m)
    return result


def summarize(vars: Sequence[AffineForm], n: int, m: int = 0) -> None:
    """In-place version of :func:`summarized`."""
    if not 2 <= len(vars) <= n:
        raise ValueError

    context = getcontext()
    keys: set[int] = set()

    for var in vars:
        if var._intvl is not vars[0]._intvl:
            raise ValueError

        var._ensurecontext(context)
        keys |= var._coeffs.keys()

    if len(keys) < min(n, m):
        return

    ZERO = vars[0]._intvl.operator.ZERO
    cadd = vars[0]._intvl.operator.cadd
    tmp: list[tuple] = []

    for key in keys:
        a0 = a1 = ZERO

        for var in vars:
            if (a := var._coeffs.get(key)) is not None:
                if abs(a) > a0:
                    a0 = abs(a)
                elif abs(a) > a1:
                    a1 = abs(a)

        penalty = a0 * a1 / (a0 + a1) if a0 != ZERO else ZERO
        tmp.append((key, penalty))

    tmp.sort(key=lambda tmp: tmp[1], reverse=True)
    del tmp[: n - len(vars)]

    for var in vars:
        rad = ZERO
        coeffs = var._coeffs

        for key, _ in tmp:
            if (a := coeffs.get(key)) is not None:
                rad = cadd(rad, abs(a))
                del coeffs[key]

        if context.rounding == "FAST":
            rad = cadd(rad, var._excess)
            var._excess = ZERO

        key = context.create_noisesymbol()
        coeffs[key] = rad



import dataclasses
import itertools
from collections.abc import Callable
from typing import Literal

from verry.autodiff.autodiff import deriv, jacobian
from verry.interval.interval import Interval
from verry.linalg.intervalmatrix import IntervalMatrix, LinAlgError, approx_inv


@dataclasses.dataclass(frozen=True, slots=True)
class AllRootResult[T: IntervalMatrix]:
    """Output of :func:`allroot`.

    Attributes
    ----------
    exists : list[IntervalMatrix]
        Interval vectors in which at least one root exists.
    unique : list[IntervalMatrix]
        Interval vectors in which exactly one root exists.
    unknown : list[IntervalMatrix]
        Interval vectors in which the existence of roots could not be verified.
    """

    exists: list[T] = dataclasses.field(default_factory=list)
    unique: list[T] = dataclasses.field(default_factory=list)
    unknown: list[T] = dataclasses.field(default_factory=list)


@dataclasses.dataclass(frozen=True, slots=True)
class AllRootScalarResult[T: Interval]:
    """Output of :func:`allroot_scalar`.

    Attributes
    ----------
    exists : list[Interval]
        Intervals in which at least one root exists.
    unique : list[Interval]
        Intervals in which exactly one root exists.
    unknown : list[Interval]
        Intervals in which the existence of roots could not be verified.
    """

    exists: list[T] = dataclasses.field(default_factory=list)
    unique: list[T] = dataclasses.field(default_factory=list)
    unknown: list[T] = dataclasses.field(default_factory=list)


def allroot[T: IntervalMatrix](
    fun: Callable,
    domain: T,
    fprime: Callable | None = None,
    unique: bool = False,
    max_iter: int = 16,
) -> AllRootScalarResult[T]:
    """Find all roots of multivariate scalar-valued function.

    Parameters
    ----------
    fun : Callable
        Function to find a root of.
    domain : IntervalMatrix
        Interval vector for which roots are searched.
    fprime : Callable, optional
        Derivative of `fun` (the default is ``jacobian(fun)``).
    unique : bool, default=False
        If `unique` is ``True``, verification continues until the number of iterations
        reaches `max_iter` or the uniqueness of the root is verified, even if its
        existence has been established.
    max_iter : int, default=16
        Maximum number of iterations.

    Returns
    -------
    AllRootResult

    Warnings
    --------
    `fun` must be a :math:`C^1`-function on `domain`. Futhermore, `fun` must neither be
    a constant nor contain conditional branches (cf. :doc:`/userguide/pitfall`).

    See Also
    --------
    verry.autodiff.jacobian

    Examples
    --------
    >>> from verry.linalg import FloatIntervalMatrix as FIM
    >>> fun = lambda x, y: (x**2 - y - 3, -x + y**2 - 3)
    >>> r = allroot(fun, FIM(inf=[-3, -3], sup=[3, 3]), unique=True)
    >>> len(r.unique)
    4
    >>> root_min = min(r.unique, key=lambda x: x[0].inf)
    >>> -2 in root_min[0] and 1 in root_min[1]
    True
    """
    if not isinstance(domain, IntervalMatrix):
        raise TypeError

    if domain.ndim != 1:
        raise ValueError

    intvlmat: type[T] = type(domain)
    intvl = domain.interval

    if fprime is None:
        fprime = jacobian(fun)

    if max_iter <= 0:
        raise ValueError

    cands = [domain]
    result: AllRootResult[T] = AllRootResult()

    for i in range(max_iter):
        if not cands:
            break

        next_cands: list[T] = []

        for x in cands:
            match krawczyk(fun, x, fprime):
                case "EXISTS", root:
                    if i == max_iter - 1 or not unique:
                        result.exists.append(root)  # type: ignore
                        continue

                    tmp = ((intvl(y.inf, y.mid()), intvl(y.mid(), y.sup)) for y in x)
                    next_cands.extend(intvlmat(x) for x in itertools.product(*tmp))

                case "NOTEXISTS", _:
                    pass

                case "UNIQUE", root:
                    result.unique.append(root)  # type: ignore

                case "UNKNOWN", _:
                    if i == max_iter - 1:
                        result.unknown.append(x)
                        continue

                    tmp = ((intvl(y.inf, y.mid()), intvl(y.mid(), y.sup)) for y in x)
                    next_cands.extend(intvlmat(x) for x in itertools.product(*tmp))

        cands = next_cands

    return result


def allroot_scalar[T: Interval](
    fun: Callable,
    domain: T,
    fprime: Callable | None = None,
    unique: bool = False,
    max_iter: int = 16,
) -> AllRootScalarResult[T]:
    """Find all roots of univariate scalar-valued function.

    Parameters
    ----------
    fun : Callable
        Function to find a root of.
    domain : Interval
        Interval for which roots are searched.
    fprime : Callable, optional
        Derivative of `fun` (the default is ``deriv(fun)``).
    unique : bool, default=False
        If `unique` is ``True``, verification continues until the number of iterations
        reaches `max_iter` or the uniqueness of the root is verified, even if its
        existence has been established.
    max_iter : int, default=16
        Maximum number of iterations.

    Returns
    -------
    AllRootScalarResult

    Warnings
    --------
    `fun` must be a :math:`C^1`-function on `domain`. Futhermore, `fun` must neither be
    a constant nor contain conditional branches (cf. :doc:`/userguide/pitfall`).

    See Also
    --------
    verry.autodiff.deriv

    Examples
    --------
    >>> from verry import FloatInterval as FI
    >>> from verry import function as vrf
    >>> r = allroot_scalar(lambda x: x**3 - 2 * x, FI(-2, 3), unique=True)
    >>> len(r.unique)
    3
    >>> root_max = max(r.unique, key=lambda x: x.sup)
    >>> root_max.issuperset(vrf.sqrt(FI(2)))
    True
    """
    if not isinstance(domain, Interval):
        raise TypeError

    intvl: type[T] = type(domain)

    if fprime is None:
        fprime = deriv(fun)

    if max_iter <= 0:
        raise ValueError

    cands = [domain]
    result = AllRootScalarResult()

    for i in range(max_iter):
        if not cands:
            break

        next_cands: list[T] = []

        for x in cands:
            match krawczyk_scalar(fun, x, fprime):
                case "EXISTS", root:
                    if i == max_iter - 1 or not unique:
                        result.exists.append(root)  # type: ignore
                        continue

                    next_cands.extend((intvl(x.inf, x.mid()), intvl(x.mid(), x.sup)))

                case "NOTEXISTS", _:
                    pass

                case "UNIQUE", root:
                    result.unique.append(root)  # type: ignore

                case "UNKNOWN", _:
                    if i == max_iter - 1:
                        result.unknown.append(x)
                        continue

                    next_cands.extend((intvl(x.inf, x.mid()), intvl(x.mid(), x.sup)))

        cands = next_cands

    return result


type _TestResult[T] = (
    tuple[Literal["EXISTS"], T]
    | tuple[Literal["NOTEXISTS"], None]
    | tuple[Literal["UNIQUE"], T]
    | tuple[Literal["UNKNOWN"], None]
)


def krawczyk[T: IntervalMatrix](
    fun: Callable, x: T, fprime: Callable | T | None = None
) -> _TestResult[T]:
    """Apply the Krawczyk test to the multivariate vector-valued function.

    Parameters
    ----------
    fun : Callable
        Function to find a root of.
    x : IntervalMatrix
        Candidate set for which a root is expected to exist.
    fprime : Callable | IntervalMatrix, optional
        Derivative of `fun`, or an image of `x` under the derivative (the default is
        ``jacobian(fun)``).

    Returns
    -------
    r0 : Literal["EXISTS", "NOTEXISTS", "UNIQUE", "UNKNOWN"]
    r1 : IntervalMatrix | None
        If `r0` is ``"EXISTS"`` or ``"UNIQUE"``, `r1` is the set containing the root;
        otherwise, `r1` is ``None``.

    Warnings
    --------
    `fun` must be a :math:`C^1`-function on `x`. Futhermore, `fun` must neither be a
    constant nor contain conditional branches (cf. :doc:`/userguide/pitfall`).

    See Also
    --------
    krawczyk_scalar, verry.autodiff.jacobian
    """
    if not isinstance(x, IntervalMatrix):
        raise TypeError

    if x.ndim != 1:
        raise ValueError

    intvlmat = type(x)
    intvl = x.interval

    match fprime:
        case Callable():  # type: ignore
            c = intvlmat(fprime(*x), intvl=intvl)

        case x.__class__():
            c = fprime

        case None:
            c = intvlmat(jacobian(fun)(*x), intvl=intvl)

        case _:
            raise TypeError

    x0 = intvlmat(x.mid(), intvl=intvl)

    if not x.interiorcontains(x0):
        return ("UNKNOWN", None)

    try:
        r = approx_inv(c)
    except LinAlgError:
        return ("UNKNOWN", None)

    tmp = intvlmat.eye(len(x), intvl=intvl) - r @ c
    x1 = x0 - r @ intvlmat(fun(*x0), intvl=intvl) + tmp @ (x - x0)

    if x.interiorcontains(x1):
        return ("UNIQUE", x1)

    if x.issuperset(x1):
        return ("EXISTS", x1)

    if x.isdisjoint(x1):
        return ("NOTEXISTS", None)

    return ("UNKNOWN", None)


def krawczyk_scalar[T: Interval](
    fun: Callable, x: T, fprime: Callable | T | None = None
) -> _TestResult[T]:
    """Apply the Krawczyk test to the univariate scalar-valued function.

    Parameters
    ----------
    fun : Callable
        Function to find a root of.
    x : Interval
        Candidate set for which a root is expected to exist.
    fprime : Callable | Interval, optional
        Derivative of `fun`, or an image of `x` under the derivative (the default is
        ``deriv(fun)``).

    Returns
    -------
    r0 : Literal["EXISTS", "NOTEXISTS", "UNIQUE", "UNKNOWN"]
    r1 : Interval | None
        If `r0` is ``"EXISTS"`` or ``"UNIQUE"``, `r1` is the set containing the root;
        otherwise, `r1` is ``None``.

    Warnings
    --------
    `fun` must be a :math:`C^1`-function on `x`. Futhermore, `fun` must neither be a
    constant nor contain conditional branches (cf. :doc:`/userguide/pitfall`).

    See Also
    --------
    krawczyk, verry.autodiff.deriv
    """
    if not isinstance(x, Interval):
        raise TypeError

    intvl = type(x)

    match fprime:
        case Callable():  # type: ignore
            c = fprime(x)

        case x.__class__():
            c = fprime

        case None:
            c = deriv(fun)(x)

        case _:
            raise TypeError

    if not isinstance(c, intvl):
        raise TypeError

    ONE = intvl.operator.ONE
    ZERO = intvl.operator.ZERO
    x0 = intvl(x.mid())

    if not x.interiorcontains(x0):
        return ("UNKNOWN", None)

    if (c0 := c.mid()) == ZERO:
        return ("UNKNOWN", None)

    r = ONE / c0
    x1 = x0 - r * fun(x0) + (1 - r * c) * (x - x0)

    if x.interiorcontains(x1):
        return ("UNIQUE", x1)

    if x.issuperset(x1):
        return ("EXISTS", x1)

    if x.isdisjoint(x1):
        return ("NOTEXISTS", None)

    return ("UNKNOWN", None)



from collections.abc import Iterator, Sequence
from typing import Self

from verry.typing import Scalar


class Vector[T: Scalar]:
    __slots__ = ("_coeffs",)
    _coeffs: list[T]

    def __init__(self, coeffs: Self | Sequence[T]):
        self._coeffs = list(coeffs)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._coeffs!r})"

    def __len__(self) -> int:
        return len(self._coeffs)

    def __iter__(self) -> Iterator[T]:
        return iter(self._coeffs)

    def __eq__(self, other) -> bool:
        if type(other) is not type(self):
            return NotImplemented

        return other._coeffs == self._coeffs

    def __getitem__(self, key: int) -> T:
        return self._coeffs[key]

    def __setitem__(self, key: int, value: T) -> None:
        self._coeffs[key] = value

    def __add__(self, rhs: Self) -> Self:
        return self.__class__([x + y for x, y in zip(self._coeffs, rhs._coeffs)])

    def __sub__(self, rhs: Self) -> Self:
        return self.__class__([x - y for x, y in zip(self._coeffs, rhs._coeffs)])

    def __mul__(self, rhs) -> Self:
        return self.__class__([x * rhs for x in self._coeffs])

    def __truediv__(self, rhs) -> Self:
        return self.__class__([x / rhs for x in self._coeffs])

    def __neg__(self) -> Self:
        return self.__class__([-x for x in self._coeffs])

    def __pos__(self) -> Self:
        return self.__class__(self._coeffs)

    def __radd__(self, lhs: Self) -> Self:
        return self.__add__(lhs)

    def __rsub__(self, lhs: Self) -> Self:
        return self.__neg__().__add__(lhs)

    def __rmul__(self, lhs) -> Self:
        return self.__mul__(lhs)


class Dual[T: Scalar](Scalar):
    __slots__ = ("real", "imag", "_level")
    real: T
    imag: Vector[T]
    _level: int

    def __init__(self, real: T, imag: Vector[T] | Sequence[T]):
        self.real = real
        self.imag = Vector(imag)
        self._level = (real._level + 1) if isinstance(real, type(self)) else 1

    @property
    def level(self) -> int:
        return self._level

    @classmethod
    def variable(cls, *args: T) -> tuple[Self, ...]:
        result: list[Self] = []

        for argnum, arg in enumerate(args):
            ZERO = arg * 0
            ONE = ZERO + 1
            imag = tuple(ONE if i == argnum else ZERO for i in range(len(args)))
            result.append(cls(arg, imag))

        return tuple(result)

    def __repr__(self) -> str:
        return f"{type(self).__name__}(real={self.real!r}, imag={self.imag._coeffs!r})"

    def __eq__(self, other) -> bool:
        if type(other) is not type(self):
            return NotImplemented

        return other.real == self.real and other.imag == self.imag

    def __add__(self, rhs) -> Self:
        if not isinstance(rhs, type(self)) or rhs._level < self._level:
            return self.__class__(self.real + rhs, self.imag)

        if rhs._level > self._level:
            return self.__class__(self + rhs.real, rhs.imag)  # type: ignore

        imag = self.imag + rhs.imag
        return self.__class__(self.real + rhs.real, imag)

    def __sub__(self, rhs) -> Self:
        if not isinstance(rhs, type(self)) or rhs._level < self._level:
            return self.__class__(self.real - rhs, self.imag)

        if rhs._level > self._level:
            return self.__class__(self - rhs.real, -rhs.imag)  # type: ignore

        imag = self.imag - rhs.imag
        return self.__class__(self.real - rhs.real, imag)

    def __mul__(self, rhs) -> Self:
        if isinstance(rhs, Vector):
            return NotImplemented

        if not isinstance(rhs, type(self)) or rhs._level < self._level:
            return self.__class__(self.real * rhs, self.imag * rhs)

        if rhs._level > self._level:
            return self.__class__(self * rhs.real, self * rhs.imag)  # type: ignore

        imag = self.real * rhs.imag + rhs.real * self.imag
        return self.__class__(self.real * rhs.real, imag)

    def __truediv__(self, rhs) -> Self:
        if not isinstance(rhs, type(self)) or rhs._level < self._level:
            return self.__class__(self.real / rhs, self.imag / rhs)

        if rhs._level > self._level:
            imag = -self * rhs.imag / rhs.real**2
            return self.__class__(self / rhs.real, imag)  # type: ignore

        imag = (self.imag * rhs.real - self.real * rhs.imag) / rhs.real**2
        return self.__class__(self.real / rhs.real, imag)

    def __pow__(self, rhs: int) -> Self:
        imag = rhs * self.real ** (rhs - 1) * self.imag
        return self.__class__(self.real**rhs, imag)

    def __neg__(self) -> Self:
        return self.__class__(-self.real, -self.imag)

    def __pos__(self) -> Self:
        return self.__class__(self.real, self.imag)

    def __radd__(self, lhs) -> Self:
        return self.__add__(lhs)

    def __rsub__(self, lhs) -> Self:
        return self.__neg__().__add__(lhs)

    def __rmul__(self, lhs) -> Self:
        return self.__mul__(lhs)

    def __rtruediv__(self, lhs) -> Self:
        if isinstance(lhs, Vector):
            return NotImplemented

        if isinstance(lhs, type(self)):
            return lhs.__truediv__(self)

        imag = -lhs * self.imag / self.real**2
        return self.__class__(lhs / self.real, imag)



"""
#################################
Miscellaneous (:mod:`verry.misc`)
#################################

This module provides some utilities, primarily for library developers.

.. autosummary::
    :toctree: generated/

    FormatSpec

"""

from .formatspec import FormatSpec

__all__ = ["FormatSpec"]



"""
#############################################
Interval series (:mod:`verry.intervalseries`)
#############################################

.. currentmodule:: verry.intervalseries

This module provides operations on power series with interval coefficients.

Interval series
===============

.. autosummary::
    :toctree: generated/

    IntervalSeries

Context
=======

.. autosummary::
    :toctree: generated/

    Context
    getcontext
    localcontext
    setcontext

"""

import contextlib
import contextvars
from collections.abc import Sequence
from typing import Any, Literal, Self, overload

from verry import function as vrf
from verry.interval.interval import Interval
from verry.typing import Scalar, SignedComparable


class Context[T: Interval]:
    """Create a new context."""

    __slots__ = ("rounding", "deg", "domain")
    rounding: Literal["TYPE1", "TYPE2"]
    deg: int
    domain: T | None

    def __init__(
        self,
        rounding: Literal["TYPE1", "TYPE2"] = "TYPE1",
        deg: int = 15,
        domain: T | None = None,
    ):
        self.rounding = rounding
        self.deg = deg
        self.domain = domain

        if rounding == "TYPE2" and domain is None:
            raise ValueError

    def copy(self) -> Self:
        return self.__class__(self.rounding, self.deg, self.domain)

    def round(self, series: "IntervalSeries") -> None:
        match self.rounding:
            case "TYPE1":
                series.round_type1(self.deg)

            case "TYPE2":
                if not isinstance(self.domain, series.interval):
                    raise TypeError

                series.round_type2(self.deg, self.domain)

    def __repr__(self):
        rounding = self.rounding
        deg = self.deg

        match rounding:
            case "TYPE1":
                return f"{type(self).__name__}({rounding=!r}, {deg=})"

            case "TYPE2":
                domain = self.domain
                return f"{type(self).__name__}({rounding=!r}, {deg=}, {domain=!r})"


_var: contextvars.ContextVar[Context] = contextvars.ContextVar("intervalseries")


def getcontext() -> Context:
    """Return the current context for the active thread."""
    if ctx := _var.get(None):
        return ctx

    ctx = Context()
    _var.set(ctx)
    return ctx


def setcontext(ctx: Context) -> None:
    """Set the current context for the active thread to `ctx`."""
    _var.set(ctx)


@contextlib.contextmanager
def localcontext(
    ctx: Context | None = None,
    *,
    rounding: Literal["TYPE1", "TYPE2"] | None = None,
    deg: int | None = None,
    domain: Interval | None = None,
):
    """Return a context manager that will set the current context for the active thread
    to a copy of `ctx` on entry to the with-statement and restore the previous context
    when exiting the with-statement."""
    if ctx is None:
        ctx = getcontext()

    if rounding is None:
        rounding = ctx.rounding

    if deg is None:
        deg = ctx.deg

    if domain is None:
        domain = ctx.domain

    ctx = Context(rounding, deg, domain)
    token = _var.set(ctx)

    try:
        yield ctx
    finally:
        _var.reset(token)


class IntervalSeries[T1: Interval, T2: SignedComparable = Any](Scalar):
    """Interval series.

    Parameters
    ----------
    coeffs
        Sequence of coefficients with respect to the monomial basis.
    intvl : type[Interval] | None
        Type of each coefficient.
    """

    __slots__ = ("coeffs", "_intvl")
    coeffs: list[T1]
    _intvl: type[T1]

    def __init__(self, coeffs: Sequence[T1 | T2 | int | float | str], intvl: type[T1]):
        if not issubclass(intvl, Interval):
            raise TypeError

        self.coeffs = [intvl.ensure(x) for x in coeffs]
        self._intvl = intvl

    @property
    def interval(self) -> type[T1]:
        return self._intvl

    def __str__(self) -> str:
        return f"{type(self).__name__}({self.coeffs})"

    def compose(self, arg: Self) -> Self:
        """Return the composition.

        See Also
        --------
        eval, __call__

        Examples
        --------
        >>> from verry import FloatInterval as FI
        >>> f = IntervalSeries([4, 3, 1], intvl=FI)
        >>> g = IntervalSeries([2, -1, 5], intvl=FI)
        >>> h = f.compose(g)
        >>> print([coeff.mid() for coeff in h.coeffs])
        [14.0, -7.0, 36.0, -10.0, 25.0]
        """
        result = self.__class__([self.coeffs[-1]], intvl=self._intvl)

        for x in reversed(self.coeffs[:-1]):
            result *= arg
            result += x

        getcontext().round(result)
        return result

    def copy(self) -> Self:
        """Return a copy of the series."""
        return self.__class__([x.copy() for x in self.coeffs], intvl=self._intvl)

    def eval(self, arg: T1 | T2 | int | float) -> T1:
        """Return an interval containing the image of `arg`.

        See Also
        --------
        compose, __call__

        Examples
        --------
        >>> from verry import FloatInterval as FI
        >>> f = IntervalSeries([8, 2, -1], intvl=FI)
        >>> y1 = f.eval(2)
        >>> print(format(y1, ".2f"))
        [8.00, 8.00]
        >>> y2 = f.eval(FI(-1, 1))
        >>> FI(5, 9).issubset(y2)
        True
        """
        result = self.coeffs[-1].copy()

        for x in reversed(self.coeffs[:-1]):
            result *= arg
            result += x

        return result

    def integrate(self) -> Self:
        """Integrate the series in a term by term.

        Examples
        --------
        >>> from verry import FloatInterval as FI
        >>> x = IntervalSeries([0, 1], intvl=FI)
        >>> y = 3 * x**2 + 4 * x + 5
        >>> z = y.integrate()
        >>> z == x**3 + 2 * x**2 + 5 * x
        True
        """
        coeffs = [self._intvl(), *(x.copy() for x in self.coeffs)]

        for i in range(2, len(coeffs)):
            coeffs[i] = coeffs[i] / i

        return self.__class__(coeffs, intvl=self._intvl)

    def reciprocal(self) -> Self:
        """Return the reciprocal of the series.

        Raises
        ------
        ZeroDivisionError
            If the constant term contains zero.
        """
        ZERO = self._intvl.operator.ZERO

        if ZERO in (x0 := self.coeffs[0]):
            raise ZeroDivisionError

        context = getcontext()
        dx = self.copy()
        dx.coeffs[0] = self._intvl()

        result = self.__class__([0], intvl=self._intvl)
        coeff = 1 / x0
        dxpow = self.__class__([1], intvl=self._intvl)

        for _ in range(1, context.deg):
            result += coeff * dxpow
            coeff /= -x0
            dxpow *= dx

        match context.rounding:
            case "TYPE1":
                result += coeff * dxpow
                return result

            case "TYPE2":
                if not isinstance(context.domain, self._intvl):
                    raise TypeError

                coeff = -pow(-1 / self.eval(x0 | context.domain), context.deg + 1)
                result += coeff * dxpow
                return result

    def round_type1(self, deg: int) -> None:
        """Round the series in Type-I PSA. This method modifies the series in-place.

        Parameters
        ----------
        deg
            Truncation degree.

        See Also
        --------
        round_type2

        Examples
        --------
        >>> from verry import FloatInterval as FI
        >>> f = IntervalSeries([1, 2, -3], intvl=FI)
        >>> g = IntervalSeries([1, -1, 1], intvl=FI)
        >>> h1 = f * g
        >>> h2 = h1.copy()
        >>> h2.round_type1(deg=2)
        >>> h1(1).issubset(h2(1))
        False
        """
        if (n := deg + 1) < len(self.coeffs):
            del self.coeffs[n:]

    def round_type2(self, deg: int, domain: T1) -> None:
        """Round the series in Type-II PSA. This method modifies the series in-place.

        Parameters
        ----------
        deg
            Truncation degree.

        See Also
        --------
        round_type1

        Examples
        --------
        >>> from verry import FloatInterval as FI
        >>> f = IntervalSeries([1, 2, -3], intvl=FI)
        >>> g = IntervalSeries([1, -1, 1], intvl=FI)
        >>> h1 = f * g
        >>> h2 = h1.copy()
        >>> h2.round_type2(deg=2, domain=FI(0, 2))
        >>> h1(1).issubset(h2(1))
        True
        """
        if not isinstance(domain, self._intvl):
            raise TypeError

        if deg >= len(self.coeffs) - 1:
            return

        tmp = self.coeffs[-1].copy()

        for x in reversed(self.coeffs[deg:-1]):
            tmp *= domain
            tmp += x

        del self.coeffs[deg:]
        self.coeffs.append(tmp)

    def _verry_overload_(self, fun, *args, **kwargs):
        if fun is vrf.e:
            return self.__class__([vrf.e(self._intvl())], intvl=self._intvl)

        if fun is vrf.ln2:
            return self.__class__([vrf.ln2(self._intvl())], intvl=self._intvl)

        if fun is vrf.pi:
            return self.__class__([vrf.pi(self._intvl())], intvl=self._intvl)

        return NotImplemented

    def __eq__(self, other) -> bool:
        if type(other) is not type(self):
            return NotImplemented

        return other._intvl is self._intvl and other.coeffs == self.coeffs

    @overload
    def __call__(self, arg: Self) -> Self: ...

    @overload
    def __call__(self, arg: T1 | T2 | int | float) -> T1: ...

    def __call__(self, arg):
        """Return ``compose(arg)`` or ``eval(arg)`` depending on the type of `arg`.

        See Also
        --------
        compose, eval
        """
        match arg:
            case self._intvl.endtype() | self._intvl() | int() | float():
                return self.eval(arg)

            case self.__class__():
                return self.compose(arg)

            case _:
                raise TypeError

    def __add__(self, rhs: Self | T1 | T2 | int | float) -> Self:
        match rhs:
            case self._intvl.endtype() | self._intvl() | int() | float():
                result = self.copy()
                result.coeffs[0] += rhs
                return result

            case self.__class__():
                n, m = len(self.coeffs), len(rhs.coeffs)
                coeffs = [self.coeffs[i] + rhs.coeffs[i] for i in range(min(n, m))]

                if n < m:
                    for i in range(n, m):
                        coeffs.append(rhs.coeffs[i])
                else:
                    for i in range(m, n):
                        coeffs.append(self.coeffs[i])

                return self.__class__(coeffs, intvl=self._intvl)

            case _:
                return NotImplemented

    def __sub__(self, rhs: Self | T1 | T2 | int | float) -> Self:
        match rhs:
            case self._intvl.endtype() | self._intvl() | int() | float():
                result = self.copy()
                result.coeffs[0] -= rhs
                return result

            case self.__class__():
                n, m = len(self.coeffs), len(rhs.coeffs)
                coeffs = [self.coeffs[i] - rhs.coeffs[i] for i in range(min(n, m))]

                if n < m:
                    for i in range(n, m):
                        coeffs.append(-rhs.coeffs[i])
                else:
                    for i in range(m, n):
                        coeffs.append(self.coeffs[i])

                return self.__class__(coeffs, intvl=self._intvl)

            case _:
                return NotImplemented

    def __mul__(self, rhs: Self | T1 | T2 | int | float) -> Self:
        match rhs:
            case self._intvl.endtype() | self._intvl() | int() | float():
                return self.__class__([x * rhs for x in self.coeffs], intvl=self._intvl)

            case self.__class__():
                n, m = len(self.coeffs), len(rhs.coeffs)
                result = self.__class__([0], intvl=self._intvl)
                result.coeffs = [self._intvl()] * (n + m - 1)

                for i in range(n):
                    for j in range(m):
                        result.coeffs[i + j] += self.coeffs[i] * rhs.coeffs[j]

                getcontext().round(result)
                return result

            case _:
                return NotImplemented

    def __truediv__(self, rhs: Self | T1 | T2 | int | float) -> Self:
        match rhs:
            case self._intvl.endtype() | self._intvl() | int() | float():
                return self.__class__([x / rhs for x in self.coeffs], intvl=self._intvl)

            case self.__class__():
                return self.__mul__(rhs.reciprocal())

            case _:
                return NotImplemented

    def __pow__(self, rhs: int) -> Self:
        if not isinstance(rhs, int):
            return NotImplemented

        if rhs < 0:
            return self.__pow__(-rhs).reciprocal()

        result = self.__class__([1], intvl=self._intvl)
        tmp = self.copy()

        while rhs != 0:
            if rhs % 2 != 0:
                result *= tmp

            rhs //= 2
            tmp *= tmp

        return result

    def __radd__(self, lhs: Self | T1 | T2 | int | float) -> Self:
        return self.__add__(lhs)

    def __rsub__(self, lhs: Self | T1 | T2 | int | float) -> Self:
        return self.__neg__().__add__(lhs)

    def __rmul__(self, lhs: Self | T1 | T2 | int | float) -> Self:
        return self.__mul__(lhs)

    def __rtruediv__(self, lhs: Self | T1 | T2 | int | float) -> Self:
        return self.reciprocal().__mul__(lhs)

    def __neg__(self) -> Self:
        return self.__class__([-x for x in self.coeffs], intvl=self._intvl)

    def __pos__(self) -> Self:
        return self.copy()

    def __copy__(self) -> Self:
        return self.copy()



import itertools
from collections.abc import Iterator, Sequence
from types import EllipsisType
from typing import Any, Literal, Self, Union, overload

import numpy as np
import numpy.typing as npt

from verry import function as vrf
from verry.interval.interval import Interval
from verry.typing import ComparableScalar


class LinAlgError(ValueError):
    """Error raised by :mod:`verry.linalg` functions."""


class flatiter[T: Interval](Iterator[T]):
    __slots__ = ("_iter", "_matrix")
    _iter: Iterator
    _matrix: "IntervalMatrix[T]"

    def __init__(self, a: "IntervalMatrix[T]", /):
        self._iter = iter(itertools.product(*(range(n) for n in a.shape)))
        self._matrix = a

    def __iter__(self) -> Self:
        return self

    def __next__(self) -> T:
        return self._matrix[*next(self._iter)]


class matiter(Iterator):
    __slots__ = ("_iter", "_matrix")
    _iter: Iterator
    _matrix: "IntervalMatrix"

    def __init__(self, a: "IntervalMatrix", /):
        self._iter = iter(range(len(a)))
        self._matrix = a

    def __iter__(self) -> Self:
        return self

    def __next__(self) -> Any:
        return self._matrix[next(self._iter)]


class IntervalMatrix[T1: Interval, T2: ComparableScalar = Any]:
    """Inf-sup type interval matrix.

    Parameters
    ----------
    a : IntervalMatrix | ndarray | Sequence
    intvl : type[Interval]

    Attributes
    ----------
    inf : ndarray
        Infimum of the interval matrix.
    sup : ndarray
        Supremum of the interval matrix.
    """

    __slots__ = ("inf", "sup", "_intvl")
    __array_ufunc__ = None
    inf: npt.NDArray
    sup: npt.NDArray
    _intvl: type[T1]

    @overload
    def __init__(
        self,
        inf: Union[
            "IntervalMatrix[T1]",
            npt.NDArray,
            Sequence[T1 | T2 | float | int | str],
            Sequence[npt.NDArray | Sequence[T1 | T2 | float | int | str]],
        ],
        /,
        *,
        intvl: type[T1],
    ): ...

    @overload
    def __init__(
        self,
        inf: npt.NDArray
        | Sequence[T2 | float | int | str]
        | Sequence[npt.NDArray | Sequence[T2 | float | int | str]],
        sup: npt.NDArray
        | Sequence[T2 | float | int | str]
        | Sequence[npt.NDArray | Sequence[T2 | float | int | str]],
        *,
        intvl: type[T1],
    ): ...

    def __init__(self, inf, sup=None, *, intvl, **kwargs):
        if not issubclass(intvl, Interval):
            raise TypeError

        self._intvl = intvl

        if kwargs.get("_skipcheck"):
            self.inf = inf
            self.sup = sup
            return

        if sup is not None:
            if not isinstance(inf, np.ndarray):
                inf = np.array(inf, np.object_)

            if not isinstance(sup, np.ndarray):
                sup = np.array(sup, np.object_)

            if not (inf.shape == sup.shape and 1 <= inf.ndim <= 2):
                raise ValueError

            self.inf = self._emptyarray(inf.shape)  # type: ignore
            self.sup = self._emptyarray(sup.shape)  # type: ignore

            for key in itertools.product(*(range(n) for n in inf.shape)):
                self.inf[key] = intvl(inf[key]).inf  # type: ignore
                self.sup[key] = intvl(sup[key]).sup  # type: ignore

                if self.inf[key] > self.sup[key]:
                    raise ValueError

            return

        if isinstance(inf, IntervalMatrix):
            if inf._intvl is not intvl:
                raise TypeError

            self.inf = self._emptyarray(inf.shape)  # type: ignore
            self.sup = self._emptyarray(inf.shape)  # type: ignore

            for key in itertools.product(*(range(n) for n in inf.shape)):
                self.inf[key] = inf[key].inf  # type: ignore
                self.sup[key] = inf[key].sup  # type: ignore

            return

        tmp = np.array(inf, np.object_)

        if not 1 <= tmp.ndim <= 2:
            raise ValueError

        self.inf = self._emptyarray(tmp.shape)  # type: ignore
        self.sup = self._emptyarray(tmp.shape)  # type: ignore

        for key in itertools.product(*(range(n) for n in tmp.shape)):
            x = intvl.ensure(tmp[key])
            self.inf[key] = x.inf
            self.sup[key] = x.sup

    @property
    def flat(self) -> flatiter[T1]:
        return flatiter(self)

    @property
    def interval(self) -> type[T1]:
        return self._intvl

    @property
    def ndim(self) -> int:
        """Shorthand for ``len(self.shape)``.

        Unlike NumPy's ndarray, ``x.ndim`` is always greater than or equal to 1, i.e.,
        there is no 0D matrix.
        """
        return len(self.shape)

    @property
    def shape(self) -> tuple[int] | tuple[int, int]:
        """Tuple of matrix dimensions.

        Examples
        --------
        >>> from verry.linalg import FloatIntervalMatrix as FIM
        >>> x = FIM([[1, 2, 3], [4, 5, 6]])
        >>> y = FIM([1.0, 2.0])
        >>> x.shape
        (2, 3)
        >>> y.shape
        (2,)
        """
        return self.inf.shape  # type: ignore

    @property
    def size(self) -> int:
        return self.inf.size

    @property
    def T(self) -> Self:
        """Shorthand for :meth:`transpose`.

        See Also
        --------
        transpose
        """
        return self.transpose()

    @classmethod
    def empty(
        cls, shape: int | tuple[int] | tuple[int, int], *, intvl: type[T1]
    ) -> Self:
        """Return a new interval matrix of given shape, without initializing entries."""
        if not issubclass(intvl, Interval):
            raise TypeError

        inf = cls._emptyarray(shape)
        sup = inf.copy()
        return cls(inf, sup, intvl=intvl, _skipcheck=True)  # type: ignore

    @classmethod
    def eye(cls, n: int, m: int | None = None, *, intvl: type[T1]) -> Self:
        """Return a matrix with ones on the diagonal and zeros elsewhere."""
        if not issubclass(intvl, Interval):
            raise TypeError

        if m is None:
            m = n

        result = cls.zeros((n, m), intvl=intvl)
        ONE = result._intvl.operator.ONE

        for i in range(min(n, m)):
            result.inf[i, i] = ONE
            result.sup[i, i] = ONE

        return result

    @classmethod
    def ones(
        cls, shape: int | tuple[int] | tuple[int, int], *, intvl: type[T1]
    ) -> Self:
        """Return a new interval matrix of given shape, filled with ones."""
        if not issubclass(intvl, Interval):
            raise TypeError

        result = cls.empty(shape, intvl=intvl)
        ONE = result._intvl.operator.ONE
        result.inf[...] = ONE
        result.sup[...] = ONE
        return result

    @classmethod
    def zeros(
        cls, shape: int | tuple[int] | tuple[int, int], *, intvl: type[T1]
    ) -> Self:
        """Return a new interval matrix of given shape, filled with zeros."""
        if not issubclass(intvl, Interval):
            raise TypeError

        result = cls.empty(shape, intvl=intvl)
        ZERO = result._intvl.operator.ZERO
        result.inf[...] = ZERO
        result.sup[...] = ZERO
        return result

    @classmethod
    def _emptyarray(cls, shape: int | tuple[int] | tuple[int, int]) -> npt.NDArray:
        return np.empty(shape, np.object_)

    def copy(self) -> Self:
        """Return a copy of the interval matrix."""
        cls, inf, sup = type(self), self.inf.copy(), self.sup.copy()
        return cls(inf, sup, intvl=self._intvl, _skipcheck=True)  # type: ignore

    def empty_like(self) -> Self:
        """Return a new interval matrix with the same shape as `self`."""
        return self.empty(self.shape, intvl=self._intvl)

    def diam(self) -> npt.NDArray:
        """Return component-wise upper bounds of the diameter."""
        result = self._emptyarray(self.shape)

        for key in itertools.product(*(range(n) for n in self.shape)):
            result[key] = self[key].diam()  # type: ignore

        return result

    def flatten(self) -> Self:
        """Return a copy of the interval matrix collapsed into one dimension."""
        cls, inf, sup = type(self), self.inf.flatten(), self.sup.flatten()
        return cls(inf, sup, intvl=self._intvl, _skipcheck=True)  # type: ignore

    def interiorcontains(self, other: Self | npt.NDArray) -> bool:
        """Return ``True`` if the interior of the interval matrix contains `other`."""
        cond1 = isinstance(other, np.ndarray)
        cond2 = type(self) is type(other) and self._intvl is other._intvl  # type: ignore

        if not (cond1 or cond2):
            raise TypeError

        if self.shape != other.shape:
            raise ValueError

        return all(x.interiorcontains(y) for x, y in zip(self.flat, other.flat))

    def isbounded(self) -> bool:
        """Return ``True`` if both `inf` and `sup` are bounded."""
        return all(x.isbounded() for x in self.flat)

    def isdisjoint(self, other: Self | npt.NDArray) -> bool:
        """Return ``True`` if the interval matrix has no elements in common with
        `other`."""
        cond1 = isinstance(other, np.ndarray)
        cond2 = type(self) is type(other) and self._intvl is other._intvl  # type: ignore

        if not (cond1 or cond2):
            raise TypeError

        if self.shape != other.shape:
            raise ValueError

        return all(x.isdisjoint(y) for x, y in zip(self.flat, other.flat))

    def issubset(self, other: Self) -> bool:
        """Test whether every element in the interval matrix is in `other`."""
        if not (type(self) is type(other) and self._intvl is other._intvl):
            return False

        if self.shape != other.shape:
            return False

        return all(x.issubset(y) for x, y in zip(self.flat, other.flat))

    def issuperset(self, other: Self) -> bool:
        """Test whether every element in `other` is in the interval matrix."""
        if not (type(self) is type(other) and self._intvl is other._intvl):
            return False

        if self.shape != other.shape:
            return False

        return all(x.issuperset(y) for x, y in zip(self.flat, other.flat))

    def mid(self) -> npt.NDArray:
        """Return an approximation of the midpoint.

        As with scalar intervals, ``x.mid() in x`` is guaranteed to be ``True``.
        """
        result = self._emptyarray(self.shape)

        for key in itertools.product(*(range(n) for n in self.shape)):
            result[key] = self[key].mid()  # type: ignore

        return result

    def ones_like(self) -> Self:
        """Return an interval matrix of ones with the same shape as `self`."""
        return self.ones(self.shape, intvl=self._intvl)

    def rad(self) -> npt.NDArray:
        """Return component-wise upper bounds of the radius."""
        result = self._emptyarray(self.shape)

        for key in itertools.product(*(range(n) for n in self.shape)):
            result[key] = self[key].rad()  # type: ignore

        return result

    def reshape(self, shape: int | tuple[int] | tuple[int, int]) -> Self:
        """Give a new shape to an interval matrix without changing its data."""
        cls = type(self)
        inf = self.inf.reshape(shape, copy=True)
        sup = self.sup.reshape(shape, copy=True)
        return cls(inf, sup, intvl=self._intvl, _skipcheck=True)  # type: ignore

    def transpose(self) -> Self:
        """Return a view of the transposed matrix."""
        cls, inf, sup = type(self), self.inf.T, self.sup.T
        return cls(inf, sup, intvl=self._intvl, _skipcheck=True)  # type: ignore

    def zeros_like(self) -> Self:
        """Return an interval matrix of zeros with the same shape as `self`."""
        return self.zeros(self.shape, intvl=self._intvl)

    def _verry_overload_(self, fun, *args, **kwargs):
        cls = type(self)

        if fun is approx_inv:
            return cls.__approx_inv(*args)

        if fun is approx_norm:
            return cls.__approx_norm(*args)

        if fun is approx_qr:
            return cls.__approx_qr(*args)

        if fun is approx_solve:
            if args[0] is not self:
                return cls(args[0], intvl=self._intvl).__approx_solve(args[1])

            return cls.__approx_solve(*args)

        if fun is inv:
            return cls.__inv(*args)

        if fun is norm:
            return cls.__norm(*args)

        if fun is solve:
            if args[0] is not self:
                return cls(args[0], intvl=self._intvl).__solve(*args[1:])

            return cls.__solve(*args)

        return NotImplemented

    def __eq__(self, other) -> bool:
        if type(other) is not type(self):
            return NotImplemented

        if other._intvl is not self._intvl:
            return False

        return bool(np.all(other.inf == self.inf) and np.all(other.sup == self.sup))

    def __len__(self) -> int:
        return len(self.inf)

    @overload
    def __getitem__(self, key: int) -> Any: ...

    @overload
    def __getitem__(self, key: tuple[int, int]) -> T1: ...

    @overload
    def __getitem__(
        self,
        key: slice
        | EllipsisType
        | tuple[int, slice]
        | tuple[slice, int]
        | tuple[slice, slice],
    ) -> Self: ...

    def __getitem__(self, key):
        cls, inf, sup = type(self), self.inf[key], self.sup[key]

        if isinstance(inf, np.ndarray):
            return cls(inf, sup, intvl=self._intvl, _skipcheck=True)  # type: ignore

        return self._intvl(inf, sup)

    @overload
    def __setitem__(
        self, key: tuple[int, int], value: T1 | T2 | float | int
    ) -> None: ...

    @overload
    def __setitem__(
        self,
        key: int
        | slice
        | EllipsisType
        | tuple[int, slice]
        | tuple[slice, int]
        | tuple[slice, slice],
        value: Self | T1 | T2 | npt.NDArray | float | int,
    ) -> None: ...

    def __setitem__(self, key, value):
        match value:
            case self.__class__():
                self.inf[key] = value.inf
                self.sup[key] = value.sup

            case np.ndarray():
                value = self.__class__(value, intvl=self._intvl)
                self.inf[key] = value.inf
                self.sup[key] = value.sup

            case self._intvl():
                self.inf[key] = value.inf
                self.sup[key] = value.sup

            case self._intvl.endtype():
                self.inf[key] = value
                self.sup[key] = value

            case int():
                value = self._intvl(value)
                self.inf[key] = value.inf
                self.sup[key] = value.sup

            case float():
                value = self._intvl.converter.fromfloat(value)
                self.inf[key] = value
                self.sup[key] = value

            case _:
                raise TypeError

    def __iter__(self) -> matiter:
        return matiter(self)

    def __contains__(
        self, item: Sequence[T2 | float | int | Sequence[T2 | float | int]]
    ) -> bool:
        item_ = np.array(item, dtype=np.object_)

        if item_.shape != self.shape:
            return False

        for key in itertools.product(*(range(n) for n in self.shape)):
            if item_[key] not in self[key]:  # type: ignore
                return False

        return True

    def __add__(self, rhs: Self | T1 | T2 | npt.NDArray | float | int) -> Self:
        return self.copy().__iadd__(rhs)

    def __sub__(self, rhs: Self | T1 | T2 | npt.NDArray | float | int) -> Self:
        return self.copy().__isub__(rhs)

    def __mul__(self, rhs: Self | T1 | T2 | npt.NDArray | float | int) -> Self:
        return self.copy().__imul__(rhs)

    def __truediv__(self, rhs: Self | T1 | T2 | npt.NDArray | float | int) -> Self:
        return self.copy().__itruediv__(rhs)

    def __matmul__(self, rhs: Self | npt.NDArray) -> Self | T1:
        lhs, lshape = self, self.shape

        if not isinstance(rhs, (np.ndarray, type(self))):
            return NotImplemented

        rshape = rhs.shape

        if len(lshape) == 1:
            if lshape[0] != rshape[0]:
                raise ValueError

            if len(rshape) == 1:
                return sum(lhs[k] * rhs[k] for k in range(lshape[0]))

            result = self.empty((rshape[1],), intvl=self._intvl)

            for j in range(rshape[1]):
                result[j] = sum(lhs[k] * rhs[k, j] for k in range(lshape[0]))

            return result

        if lshape[1] != rshape[0]:
            raise ValueError

        if len(rshape) == 1:
            result = self.empty((lshape[0],), intvl=self._intvl)

            for i in range(lshape[0]):
                result[i] = sum(lhs[i, k] * rhs[k] for k in range(lshape[1]))

            return result

        result = self.empty((lshape[0], rshape[1]), intvl=self._intvl)

        for i in range(lshape[0]):
            for j in range(rshape[1]):
                result[i, j] = sum(lhs[i, k] * rhs[k, j] for k in range(lshape[1]))

        return result

    def __radd__(self, lhs: Self | T1 | T2 | npt.NDArray | float | int) -> Self:
        return self.copy().__iadd__(lhs)

    def __rsub__(self, lhs: Self | T1 | T2 | npt.NDArray | float | int) -> Self:
        return self.__neg__().__iadd__(lhs)

    def __rmul__(self, lhs: Self | T1 | T2 | npt.NDArray | float | int) -> Self:
        return self.copy().__imul__(lhs)

    def __rtruediv__(self, lhs: Self | T1 | T2 | npt.NDArray | float | int) -> Self:
        match lhs:
            case self._intvl.endtype() | self._intvl() | float() | int():
                for key in itertools.product(*(range(n) for n in self.shape)):
                    self[key] = lhs / self[key]  # type: ignore

                return self

            case np.ndarray() | self.__class__():
                if self.shape != lhs.shape:
                    raise ValueError

            case _:
                return NotImplemented

        result = self.empty(self.shape, intvl=self._intvl)

        for key in itertools.product(*(range(n) for n in self.shape)):
            result[key] = lhs[key] / self[key]  # type: ignore

        return result

    def __rmatmul__(self, lhs: Self | npt.NDArray) -> Self | T1:
        rhs, rshape = self, self.shape

        if not isinstance(lhs, (np.ndarray, type(self))):
            return NotImplemented

        lshape = lhs.shape

        if len(lshape) == 1:
            if lshape[0] != rshape[0]:
                raise ValueError

            if len(rshape) == 1:
                return sum(lhs[k] * rhs[k] for k in range(lshape[0]))

            result = self.empty((rshape[1],), intvl=self._intvl)

            for j in range(rshape[1]):
                result[j] = sum(lhs[k] * rhs[k, j] for k in range(lshape[0]))

            return result

        if lshape[1] != rshape[0]:
            raise ValueError

        if len(rshape) == 1:
            result = self.empty((lshape[0],), intvl=self._intvl)

            for i in range(lshape[0]):
                result[i] = sum(lhs[i, k] * rhs[k] for k in range(lshape[1]))

            return result

        result = self.empty((lshape[0], rshape[1]), intvl=self._intvl)

        for i in range(lshape[0]):
            for j in range(rshape[1]):
                result[i, j] = sum(lhs[i, k] * rhs[k, j] for k in range(lshape[1]))

        return result

    def __iadd__(self, rhs: Self | T1 | T2 | npt.NDArray | float | int) -> Self:
        match rhs:
            case self._intvl.endtype() | self._intvl() | float() | int():
                for key in itertools.product(*(range(n) for n in self.shape)):
                    self[key] += rhs  # type: ignore

                return self

            case np.ndarray() | self.__class__():
                if self.shape != rhs.shape:
                    raise ValueError

            case _:
                return NotImplemented

        for key in itertools.product(*(range(n) for n in self.shape)):
            self[key] += rhs[key]  # type: ignore

        return self

    def __isub__(self, rhs: Self | T1 | T2 | npt.NDArray | float | int) -> Self:
        match rhs:
            case self._intvl.endtype() | self._intvl() | float() | int():
                for key in itertools.product(*(range(n) for n in self.shape)):
                    self[key] -= rhs  # type: ignore

                return self

            case np.ndarray() | self.__class__():
                if self.shape != rhs.shape:
                    raise ValueError

            case _:
                return NotImplemented

        for key in itertools.product(*(range(n) for n in self.shape)):
            self[key] -= rhs[key]  # type: ignore

        return self

    def __imul__(self, rhs: Self | T1 | T2 | npt.NDArray | float | int) -> Self:
        match rhs:
            case self._intvl.endtype() | self._intvl() | float() | int():
                for key in itertools.product(*(range(n) for n in self.shape)):
                    self[key] *= rhs  # type: ignore

                return self

            case np.ndarray() | self.__class__():
                if self.shape != rhs.shape:
                    raise ValueError

            case _:
                return NotImplemented

        for key in itertools.product(*(range(n) for n in self.shape)):
            self[key] *= rhs[key]  # type: ignore

        return self

    def __itruediv__(self, rhs: Self | T1 | T2 | npt.NDArray | float | int) -> Self:
        match rhs:
            case self._intvl.endtype() | self._intvl() | float() | int():
                for key in itertools.product(*(range(n) for n in self.shape)):
                    self[key] /= rhs  # type: ignore

                return self

            case np.ndarray() | self.__class__():
                if self.shape != rhs.shape:
                    raise ValueError

            case _:
                return NotImplemented

        for key in itertools.product(*(range(n) for n in self.shape)):
            self[key] /= rhs[key]  # type: ignore

        return self

    def __neg__(self) -> Self:
        return self.copy().__imul__(-self._intvl.operator.ONE)

    def __pos__(self) -> Self:
        return self.copy()

    def __abs__(self) -> Self:
        result = self.empty(self.shape, intvl=self._intvl)

        for key in itertools.product(*(range(n) for n in self.shape)):
            result[key] = abs(self[key])  # type: ignore

        return result

    def __copy__(self) -> Self:
        return self.copy()

    def __approx_inv(self):
        if not (len(self.shape) == 2 and self.shape[0] == self.shape[1]):
            raise LinAlgError("non-square matrix")

        ZERO = self._intvl.operator.ZERO
        ONE = self._intvl.operator.ONE
        a = self.mid()
        b = np.full_like(a, ZERO)
        n = self.shape[0]

        for i in range(n):
            b[i, i] = ONE

        for k in range(n):
            if (p := np.argmax(abs(a[k:n, k])) + k) != k:
                a[(k, p),] = a[(p, k),]
                b[(k, p),] = b[(p, k),]

            if a[k, k] == ZERO:
                raise LinAlgError("numerically singular matrix")

            for i in range(k + 1, n):
                tmp = a[i, k] / a[k, k]
                a[i, k:] -= tmp * a[k, k:]
                b[i] -= tmp * b[k]

        for i in reversed(range(n)):
            b[i] -= a[i, i + 1 :] @ b[i + 1 :]
            b[i] /= a[i, i]

        return b

    def __approx_norm(self, ord):
        if len(self.shape) == 1:
            match ord:
                case "fro":
                    raise LinAlgError

                case "inf":
                    return max(abs(x.mid()) for x in self)

                case "one":
                    return sum(abs(x.mid()) for x in self)

                case "two":
                    return vrf.sqrt(sum(x.mid() ** 2 for x in self))

                case _:
                    raise ValueError

        match ord:
            case "fro":
                return vrf.sqrt(sum(x.mid() ** 2 for x in self.flat))

            case "inf":
                n = self.shape[0]
                return max(sum(abs(x.mid()) for x in self[i, :]) for i in range(n))

            case "one":
                m = self.shape[1]
                return max(sum(abs(x.mid()) for x in self[:, j]) for j in range(m))

            case "two":
                return NotImplemented

            case _:
                raise ValueError

    def __approx_qr(self):
        ZERO = self._intvl.operator.ZERO
        ONE = self._intvl.operator.ONE
        a = self.mid()
        n, m = self.shape
        q = np.full((n, n), ZERO, like=a)

        for i in range(n):
            q[i, i] = ONE

        for k in range(m):
            tmp = a[k:, k]
            e = np.full_like(tmp, ZERO)
            e[0] = vrf.sqrt(sum(x**2 for x in tmp)) * (1 if tmp[0] >= ZERO else -1)
            v = tmp + e
            v /= vrf.sqrt(sum(x**2 for x in v))
            h = np.full((n, n), ZERO, like=a)

            for i in range(n):
                h[i, i] = ONE

            for i in range(k, n):
                for j in range(k, n):
                    h[i, j] -= 2 * v[i - k] * v[j - k]

            q = q @ h
            a = h @ a

        return (q, a)

    def __approx_solve(self, b):
        if not (len(self.shape) == 2 and self.shape[0] == self.shape[1]):
            raise LinAlgError("non-square matrix")

        if not (b.ndim == 1 and len(b) == len(self)):
            raise LinAlgError("dimension mismatch")

        ZERO = self._intvl.operator.ZERO
        a = self.mid()
        b = b.copy()
        n = len(a)

        for k in range(n):
            if (p := np.argmax(abs(a[k:n, k])) + k) != k:
                a[(k, p),] = a[(p, k),]
                b[(k, p),] = b[(p, k),]

            if a[k, k] == ZERO:
                raise LinAlgError("numerically singular matrix")

            for i in range(k + 1, n):
                tmp = a[i, k] / a[k, k]
                a[i, k:] -= tmp * a[k, k:]
                b[i] -= tmp * b[k]

        for i in reversed(range(n)):
            b[i] -= sum(a[i, i + 1 :] * b[i + 1 :])
            b[i] /= a[i, i]

        return b

    def __inv(self, r):
        if r is None:
            r = approx_inv(self)

        eye = self.eye(len(self), intvl=self._intvl)
        result = self.empty_like()

        for i in range(len(self)):
            result[:, i] = solve(self, eye[:, i], r)

        return result

    def __norm(self, ord):
        if len(self.shape) == 1:
            match ord:
                case "fro":
                    raise LinAlgError

                case "inf":
                    result = self._intvl()

                    for i in range(self.shape[0]):
                        tmp = abs(self[i])
                        result.inf = max(result.inf, tmp.inf)
                        result.sup = max(result.sup, tmp.sup)

                    return result

                case "one":
                    return sum(abs(self[i]) for i in range(self.shape[0]))

                case "two":
                    return vrf.sqrt(sum(self[i] ** 2 for i in range(self.shape[0])))

                case _:
                    raise ValueError

        match ord:
            case "fro":
                result = self._intvl()

                for i in range(self.shape[0]):
                    for j in range(self.shape[1]):
                        result += self[i, j] ** 2

                return vrf.sqrt(result)

            case "inf":
                result = self._intvl()

                for i in range(self.shape[0]):
                    tmp = sum(abs(self[i, j]) for j in range(self.shape[1]))
                    result.inf = max(result.inf, tmp.inf)
                    result.sup = max(result.sup, tmp.sup)

                return result

            case "one":
                result = self._intvl()

                for j in range(self.shape[1]):
                    tmp = sum(abs(self[i, j]) for i in range(self.shape[0]))
                    result.inf = max(result.inf, tmp.inf)
                    result.sup = max(result.sup, tmp.sup)

                return result

            case "two":
                return NotImplemented

            case _:
                raise TypeError

    def __solve(self, b, r):
        if r is None:
            r = approx_inv(self)

        if not isinstance(b, type(self)):
            b = self.__class__(b, intvl=self._intvl)

        tmp = norm(self.eye(len(self), intvl=self._intvl) - r @ self, "inf")

        if tmp.inf >= self._intvl.operator.ONE:
            raise LinAlgError("numerically singular matrix")

        mid = r @ b
        rad = norm(r @ (b - self @ mid), "inf") / (1 - tmp)
        return mid + rad * self._intvl(-1, 1) * b.ones_like()


def approx_inv(a: IntervalMatrix) -> npt.NDArray:
    """Approximately compute the inverse of an interval matrix.

    Parameters
    ----------
    a : IntervalMatrix
        Matrix to be inverted.

    Returns
    -------
    ndarray

    Raises
    ------
    LinAlgError
        If `a` is not square or inversion fails.
    """
    if (res := a._verry_overload_(approx_inv, a)) is not NotImplemented:
        return res

    raise RuntimeError


def approx_norm(a: IntervalMatrix, ord: Literal["fro", "inf", "one", "two"]) -> Any:
    """Compute a matrix or vector norm approximately.

    Parameters
    ----------
    ord : Literal["fro", "inf", "one", "two"]
        Order of the norm (see NumPy documentation).
    """
    if (res := a._verry_overload_(approx_norm, a, ord)) is not NotImplemented:
        return res

    raise RuntimeError


def approx_qr(a: IntervalMatrix) -> tuple[npt.NDArray, npt.NDArray]:
    """Compute the QR factorization of a matrix approximately.

    Parameters
    ----------
    a : IntervalMatrix
        Matrix to be QR factorized.

    Returns
    -------
    q : ndarray
        Approximately orthogonal matrix.
    r : ndarray
        Upper-triangular matrix.
    """
    if (res := a._verry_overload_(approx_qr, a)) is not NotImplemented:
        return res

    raise RuntimeError


@overload
def approx_solve(a: npt.NDArray, b: IntervalMatrix) -> npt.NDArray: ...


@overload
def approx_solve(a: IntervalMatrix, b: npt.NDArray) -> npt.NDArray: ...


def approx_solve(a, b):
    """Solve a linear equation ``a @ x = b`` approximately.

    Parameters
    ----------
    a : IntervalMatrix | ndarray
        Coefficient matrix.
    b : IntervalMatrix | ndarray
        Right-hand side of the equation.

    Returns
    -------
    ndarray

    Raises
    ------
    LinAlgError
        If `a` is numerically singular or not square.
    """
    linearized = (a, b)

    if type(a) is not type(b) and issubclass(type(b), type(a)):
        linearized = (b, a)

    for x in linearized:
        if fun := getattr(type(x), "_verry_overload_", None):
            if (res := fun(x, approx_solve, a, b)) is not NotImplemented:
                return res

    raise RuntimeError


def inv[T: IntervalMatrix](a: T, r: npt.NDArray | None = None) -> T:
    """Compute the inverse of an interval matrix.

    Parameters
    ----------
    a : IntervalMatrix
        Matrix to be inverted.
    r : ndarray, optional
        Approximate inverse of `a`.

    Returns
    -------
    IntervalMatrix

    Raises
    ------
    LinAlgError
        If `a` is not square or inversion fails.
    """
    if (res := a._verry_overload_(inv, a, r)) is not NotImplemented:
        return res

    raise RuntimeError


def norm[T: Interval](
    a: IntervalMatrix[T], ord: Literal["fro", "inf", "one", "two"]
) -> T:
    """Compute a matrix or vector norm.

    Parameters
    ----------
    ord : Literal["fro", "inf", "one", "two"]
        Order of the norm (see NumPy documentation).

    Returns
    -------
    Interval
    """
    if (res := a._verry_overload_(norm, a, ord)) is not NotImplemented:
        return res

    raise RuntimeError


@overload
def solve[T: IntervalMatrix](
    a: T, b: npt.NDArray, r: npt.NDArray | None = ...
) -> T: ...


@overload
def solve[T: IntervalMatrix](
    a: npt.NDArray, b: T, r: npt.NDArray | None = ...
) -> T: ...


def solve(a, b, r=None):
    """Solve a linear equation ``a @ x = b``.

    Parameters
    ----------
    a : IntervalMatrix | ndarray
        Coefficient matrix.
    b : IntervalMatrix | ndarray
        Right-hand side of the equation.
    r : ndarray, optional
        Approximate inverse of `a`.

    Returns
    -------
    IntervalMatrix

    Raises
    ------
    LinAlgError
        If `a` is numerically singular or not square.
    """
    linearized = (a, b)

    if type(a) is not type(b) and issubclass(type(b), type(a)):
        linearized = (b, a)

    for x in linearized:
        if fun := getattr(type(x), "_verry_overload_", None):
            if (res := fun(x, solve, a, b, r)) is not NotImplemented:
                return res

    raise RuntimeError



"""
#################################################
Automatic differentiation (:mod:`verry.autodiff`)
#################################################

.. currentmodule:: verry.autodiff

This module provides forward-mode automatic differentiation.

.. autosummary::
    :toctree: generated/

    deriv
    grad
    jacobian

"""

from .autodiff import deriv, grad, jacobian

__all__ = ["deriv", "grad", "jacobian"]
